import copy
import itertools
import warnings
from collections.abc import MutableSequence

import numpy as np
import pandas as pd
import symengine
import sympy
import sympy.stats as stats
from sympy.stats.crv_types import ExponentialDistribution, NormalDistribution
from sympy.stats.joint_rv_types import MultivariateNormalDistribution

import pharmpy.math
import pharmpy.unicode as unicode
from pharmpy.symbols import symbol


def _create_rng(seed=None):
    """Create a new random number generator"""
    if isinstance(seed, np.random.Generator):
        rng = seed
    else:
        rng = np.random.default_rng(seed)
    return rng


class RandomVariable:
    """A single random variable

    Parameters
    ----------
    name : str
        Name of the random variable
    level : str
        Name of the variability level. The default levels are IIV, IOV and RUV
    sympy_rv : sympy.RandomSymbol
        RandomSymbol to use for this random variable. See also the normal
        and joint_normal classmethods.

    Examples
    --------

    >>> import sympy
    >>> import sympy.stats
    >>> from pharmpy import RandomVariable
    >>> name = "ETA(1)"
    >>> sd = sympy.sqrt(sympy.Symbol('OMEGA(1,1)'))
    >>> rv = RandomVariable(name, "IIV", sympy.stats.Normal(name, 0, sd))
    >>> rv
    ETA(1) ~ 𝒩 (0, OMEGA(1,1))

    See Also
    --------
    normal, joint_normal
    """

    def __init__(self, name, level, sympy_rv=None):
        level = RandomVariable._canonicalize_level(level)
        self._name = name
        self.level = level
        self.symbol = symbol(name)
        self._sympy_rv = sympy_rv
        if sympy_rv is not None:
            if isinstance(sympy_rv.pspace.distribution, NormalDistribution):
                self._mean = sympy.Matrix([sympy_rv.pspace.distribution.mean])
                self._variance = sympy.Matrix([sympy_rv.pspace.distribution.std ** 2])
            elif isinstance(sympy_rv.pspace.distribution, MultivariateNormalDistribution):
                raise ValueError(
                    "Cannot create multivariate random variables using constructor. "
                    "Use the joint_normal classmethod instead."
                )
            else:
                self._mean = None
                self._variance = None
        else:
            self._mean = None
            self._variance = None
        self._symengine_variance = None
        self._joint_names = None

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.level == other.level
            and self._mean == other._mean
            and self._variance == other._variance
            and self._sympy_rv == other._sympy_rv
        )

    @staticmethod
    def _canonicalize_level(level):
        supported = ('IIV', 'IOV', 'RUV')
        ulevel = level.upper()
        if ulevel not in supported:
            raise ValueError(f'Unknown variability level {level}. Must be one of {supported}.')
        return ulevel

    @classmethod
    def normal(cls, name, level, mean, variance):
        """Create a normally distributed random variable

        Parameters
        ----------
        name : str
            Name of the random variable
        level : str
            Name of the variability level
        mean : expression or number
            Mean of the random variable
        variance : expression or number
            Variance of the random variable

        Example
        -------
        >>> from pharmpy import RandomVariable, Parameter
        >>> omega = Parameter('OMEGA_CL', 0.1)
        >>> rv = RandomVariable.normal("IIV_CL", "IIV", 0, omega.symbol)
        >>> rv
        IIV_CL ~ 𝒩 (0, OMEGA_CL)

        """
        rv = cls(name, level)
        rv._mean = sympy.Matrix([sympy.sympify(mean)])
        rv._variance = sympy.Matrix([sympy.sympify(variance)])
        if rv._variance.is_positive_semidefinite is False:
            raise ValueError(f"Mean cannot be {mean} must be positive")
        rv._symengine_variance = symengine.sympify(rv._variance)
        return rv

    @classmethod
    def joint_normal(cls, names, level, mu, sigma):
        """Create joint normally distributed random variables

        Parameters
        ----------
        names : list
            Names of the random variables
        level : str
            Variability level
        mu : matrix or list
            Vector of the means of the random variables
        sigma : matrix or list of lists
            Covariance matrix of the random variables

        Example
        -------
        >>> from pharmpy import RandomVariable, Parameter
        >>> omega_cl = Parameter("OMEGA_CL", 0.1)
        >>> omega_v = Parameter("OMEGA_V", 0.1)
        >>> corr_cl_v = Parameter("OMEGA_CL_V", 0.01)
        >>> rv1, rv2 = RandomVariable.joint_normal(["IIV_CL", "IIV_V"], 'IIV', [0, 0],
        ...     [[omega_cl.symbol, corr_cl_v.symbol], [corr_cl_v.symbol, omega_v.symbol]])
        >>> rv1
        ⎡IIV_CL⎤     ⎧⎡0⎤  ⎡ OMEGA_CL   OMEGA_CL_V⎤⎫
        ⎢      ⎥ ~ 𝒩 ⎪⎢ ⎥, ⎢                      ⎥⎪
        ⎣IIV_V ⎦     ⎩⎣0⎦  ⎣OMEGA_CL_V   OMEGA_V  ⎦⎭

        """

        mean = sympy.Matrix(mu)
        variance = sympy.Matrix(sigma)
        if variance.is_positive_semidefinite is False:
            raise ValueError('Sigma matrix is not positive semidefinite')
        rvs = []
        for name in names:
            rv = cls(name, level)
            rv._mean = mean.copy()
            rv._variance = variance.copy()
            rv._symengine_variance = symengine.Matrix(variance.rows, variance.cols, sigma)
            rv._joint_names = names.copy()
            rvs.append(rv)
        return rvs

    @property
    def name(self):
        """Name of the random variable"""
        return self._name

    @name.setter
    def name(self, name):
        if self._joint_names:
            index = self._joint_names.index(self._name)
            self._joint_names[index] = name
        self._name = name

    @property
    def joint_names(self):
        """Names of all (including this) jointly varying rvs in a list"""
        return [] if not self._joint_names else self._joint_names

    @property
    def sympy_rv(self):
        """Corresponding sympy random variable"""
        if self._sympy_rv is None:
            # Normal distribution that might have 0 variance
            if len(self._variance) == 1 and self._variance[0].is_zero:
                return sympy.Integer(0)
            elif self._mean.rows > 1:
                return sympy.stats.Normal('X', self._mean, self._variance)
            else:
                return sympy.stats.Normal(self.name, self._mean[0], sympy.sqrt(self._variance[0]))
        else:
            return self._sympy_rv

    @property
    def free_symbols(self):
        """Free symbols including random variable itself"""
        if self._mean is not None:
            return {self.symbol} | self._mean.free_symbols | self._variance.free_symbols
        else:
            free = {s for s in self.sympy_rv.pspace.free_symbols if s.name != self.name}
            return free | {self.symbol}

    @property
    def parameter_names(self):
        """List of names of all parameters used in definition"""
        if self._mean is not None:
            params = self._mean.free_symbols | self._variance.free_symbols
        else:
            params = {s for s in self.sympy_rv.pspace.free_symbols if s.name != self.name}
        return sorted([p.name for p in params])

    def subs(self, d):
        """Substitute expressions

        Parameters
        ----------
        d : dict
            Dictionary of from: to pairs for substitution

        Examples
        --------
        >>> import sympy
        >>> from pharmpy import RandomVariable, Parameter
        >>> omega = Parameter("OMEGA_CL", 0.1)
        >>> rv = RandomVariable.normal("IIV_CL", "IIV", 0, omega.symbol)
        >>> rv.subs({omega.symbol: sympy.Symbol("OMEGA_NEW")})
        >>> rv
        IIV_CL ~ 𝒩 (0, OMEGA_NEW)

        """
        if self._mean is not None:
            self._mean = self._mean.subs(d)
            self._variance = self._variance.subs(d)
            self._symengine_variance = symengine.Matrix(
                self._variance.rows, self._variance.cols, self._variance
            )
        if self._sympy_rv is not None:
            self._sympy_rv = self._sympy_rv.subs(d)

    def copy(self):
        """Make copy of RandomVariable"""
        return copy.deepcopy(self)

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memo):
        # Custom copier because symengine objects cannot be copied
        new = RandomVariable(self.name, self.level)
        new._mean = self._mean.copy()
        new._variance = self._variance.copy()
        new._symengine_variance = symengine.sympify(self._variance)
        new._sympy_rv = self._sympy_rv
        if self._joint_names is None:
            new._joint_names = None
        else:
            new._joint_names = self._joint_names.copy()
        return new

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_symengine_variance']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._symengine_variance = symengine.sympify(self._variance)

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        if self._mean is not None:  # Normal distribution
            if self._mean.rows > 1:
                name_vector = sympy.Matrix(self._joint_names)
                name_strings = sympy.pretty(name_vector, wrap_line=False, use_unicode=True).split(
                    '\n'
                )
                mu_strings = sympy.pretty(self._mean, wrap_line=False, use_unicode=True).split('\n')
                sigma_strings = sympy.pretty(
                    self._variance, wrap_line=False, use_unicode=True
                ).split('\n')
                mu_height = len(mu_strings)
                sigma_height = len(sigma_strings)
                max_height = max(mu_height, sigma_height)

                left_parens = unicode.left_parens(max_height)
                right_parens = unicode.right_parens(max_height)

                # Pad the smaller of the matrices
                if mu_height != sigma_height:
                    to_pad = mu_strings if mu_height < sigma_height else sigma_strings
                    num_lines = abs(mu_height - sigma_height)
                    padding = ' ' * len(to_pad[0])
                    for i in range(0, num_lines):
                        if i % 2 == 0:
                            to_pad.append(padding)
                        else:
                            to_pad.insert(0, padding)

                # Pad names
                if len(name_strings) < max_height:
                    num_lines = abs(max_height - len(name_strings))
                    padding = ' ' * len(name_strings[0])
                    for i in range(0, num_lines):
                        if i % 2 == 0:
                            name_strings.append(padding)
                        else:
                            name_strings.insert(0, padding)

                central_index = max_height // 2
                res = []
                enumerator = enumerate(
                    zip(name_strings, left_parens, mu_strings, sigma_strings, right_parens)
                )
                for i, (name_line, lpar, mu_line, sigma_line, rpar) in enumerator:
                    if i == central_index:
                        res.append(
                            name_line
                            + f' ~ {unicode.mathematical_script_capital_n}'
                            + lpar
                            + mu_line
                            + ', '
                            + sigma_line
                            + rpar
                        )
                    else:
                        res.append(name_line + '     ' + lpar + mu_line + '  ' + sigma_line + rpar)
                return '\n'.join(res)
            else:
                return (
                    f'{sympy.pretty(self.symbol, wrap_line=False, use_unicode=True)}'
                    f' ~ {unicode.mathematical_script_capital_n}'
                    f'({sympy.pretty(self._mean[0], wrap_line=False, use_unicode=True)}, '
                    f'{sympy.pretty(self._variance[0], wrap_line=False, use_unicode=True)})'
                )
        else:
            if isinstance(self.sympy_rv.pspace.distribution, ExponentialDistribution):
                return (
                    f'{sympy.pretty(self.symbol, use_unicode=True)} ~ '
                    f'Exp({self.sympy_rv.pspace.distribution.rate})'
                )
            else:
                return f'{sympy.pretty(self.symbol, use_unicode=True)} ~ UnknownDistribution'

    def _latex_string(self, aligned=False):
        if aligned:
            align_str = ' & '
        else:
            align_str = ''
        if self._mean.rows > 1:
            rv_vec = sympy.Matrix(self._joint_names)._repr_latex_()[1:-1]
            mean_vec = self._mean._repr_latex_()[1:-1]
            sigma = self._variance._repr_latex_()[1:-1]
            latex = (
                rv_vec
                + align_str
                + r'\sim \mathcal{N} \left('
                + mean_vec
                + ','
                + sigma
                + r'\right)'
            )
        else:
            rv = self.symbol._repr_latex_()[1:-1]
            mean = self._mean[0]._repr_latex_()[1:-1]
            sigma = (self._variance[0])._repr_latex_()[1:-1]
            latex = rv + align_str + r'\sim  \mathcal{N} \left(' + mean + ',' + sigma + r'\right)'
        if not aligned:
            latex = '$' + latex + '$'
        return latex

    def _repr_latex_(self):
        return self._latex_string()


class VariabilityLevel:
    """A variability level

    Parameters
    ----------
    name : str
        A unique identifying name
    level : int
        Numeric level. 0 is the base level. Lower levels consists of groups of higher levels.
        If for example 0 is IIV then IOV could be 1 and COUNTRY could be -1
    group : str
        Name of data column to group this level. None for no grouping (default)
    """

    def __init__(self, name, level, group=None):
        self.name = name
        self.level = level
        self.group = group


class VariabilityHierarchy:
    """Description of a variability hierarchy"""

    def __init__(self):
        self._levels = []

    @property
    def names(self):
        """Names of all variability levels"""
        return [varlev.name for varlev in self._levels]

    @property
    def levels(self):
        """All numerical levels"""
        return [varlev.level for varlev in self._levels]

    def get_name(self, i):
        """Retrieve name of variability level

        Parameters
        ----------
        i - int
            Numeric variability level

        Examples
        --------
        >>> from pharmpy import VariabilityHierarchy
        >>> hierarchy = VariabilityHierarchy()
        >>> hierarchy.add_variability_level("IIV", 0, "ID")
        >>> hierarchy.add_variability_level("IOV", 1, "OCC")
        >>> hierarchy.get_name(1)
        'IOV'
        >>> hierarchy.get_name(0)
        'IIV'

        """
        for varlev in self._levels:
            if varlev.level == i:
                return varlev.name
        raise KeyError(f'No variability level {i}')

    def add_variability_level(self, name, level, group):
        """Add variability level to hierarchy

        Parameters
        ----------
        name : str
            A unique identifying name
        level : int
            Numeric level. 0 is the base level. Lower levels consists of groups of higher levels.
            If for example 0 is IIV then IOV could be 1 and COUNTRY could be -1
        group : str
            Name of data column to group this level. None for no grouping (default)

        Examples
        --------
        >>> from pharmpy import VariabilityHierarchy
        >>> hierarchy = VariabilityHierarchy()
        >>> hierarchy.add_variability_level("IIV", 0, "ID")
        >>> hierarchy.add_variability_level("IOV", 1, "OCC")

        """
        nums = self.levels
        new = VariabilityLevel(name, level, group)
        if nums:
            if not (level == min(nums) - 1 or level == max(nums) + 1):
                raise ValueError(
                    f'Cannot set variability level {level}. '
                    'New variability level must be one level higher or one level lower '
                    'than any current level'
                )
            if level == min(nums) - 1:
                self._levels.insert(0, new)
            else:
                self._levels.append(new)
        else:
            self._levels.append(new)

    def add_higher_level(self, name, group):
        """Add a higher variability level to hierarchy

        Parameters
        ----------
        name : str
            Name of new variability level
        group : str
            Name of data column to group this level. None for no grouping (default)

        Examples
        --------
        >>> from pharmpy import VariabilityHierarchy
        >>> hierarchy = VariabilityHierarchy()
        >>> hierarchy.add_variability_level("IIV", 0, "ID")
        >>> hierarchy.add_higher_level("IOV", "OCC")

        See Also
        --------
        add_variability_level, add_lower_level
        """
        nums = self.levels
        level = max(nums) + 1
        self.add_variability_level(name, level, group)

    def add_lower_level(self, name, group):
        """Add a lower variability level to hierarchy

        Parameters
        ----------
        name : str
            Name of new variability level
        group : str
            Name of data column to group this level. None for no grouping (default)

        Examples
        --------
        >>> hierarchy = VariabilityHierarchy()
        >>> hierarchy.add_variability_level("IIV", 0, "ID")
        >>> hierarchy.add_lower_level("ICV", "COUNTRY")

        See Also
        --------
        add_variability_level, add_higher_level
        """

        nums = [varlev.level for varlev in self._levels]
        level = min(nums) - 1
        self.add_variability_level(name, level, group)

    def set_variability_level(self, level, name, group):
        """Change the name and group of variability level

        Parameters
        ----------
        level : int
            Numeric level to change
        name : str
            Name of new variability level
        group : str
            Name of data column to group this level. None for no grouping (default)

        Examples
        --------
        >>> from pharmpy import VariabilityHierarchy
        >>> hierarchy = VariabilityHierarchy()
        >>> hierarchy.add_variability_level("IIV", 0, "ID")
        >>> hierarchy.add_lower_level("ICV", "COUNTRY")
        >>> hierarchy.set_variability_level(-1, "ICV", "CENTER")

        See Also
        --------
        remove_variability_level
        """
        for varlev in self._levels:
            if varlev.level == level:
                varlev.name = name
                varlev.group = group
                break
        else:
            raise KeyError(f'No variability level {level}')

    def remove_variability_level(self, ind):
        """Remove a variability level

        Parameters
        ----------
        ind : str or int
            name or number of variability level

        Examples
        --------
        >>> from pharmpy import VariabilityHierarchy
        >>> hierarchy = VariabilityHierarchy()
        >>> hierarchy.add_variability_level("IIV", 0, "ID")
        >>> hierarchy.add_lower_level("ICV", "COUNTRY")
        >>> hierarchy.remove_variability_level(-1)

        See Also
        --------
        set_variability_level

        """
        for i, varlev in enumerate(self._levels):
            if (
                isinstance(ind, str)
                and varlev.name == ind
                or isinstance(ind, int)
                and varlev.level == ind
            ):
                index = i
                break
        else:
            raise KeyError(f'No variability level {ind}')
        n = self._levels[index].level
        if n == 0:
            raise ValueError('Cannot remove the base variability level (0)')
        del self._levels[index]
        for varlev in self._levels:
            if n < 0 and varlev.level < n:
                varlev.level += 1
            elif n > 0 and varlev.level > n:
                varlev.level -= 1

    def __len__(self):
        return len(self._levels)


class RandomVariables(MutableSequence):
    """A collection of random variables

    This class provides a container for random variables that preserves their order
    and acts list-like while also allowing for indexing on names.

    Each RandomVariables object has two VariabilityHierarchies that describes the
    allowed variability levels for the contined random variables. One hierarchy
    for residual error variability (epsilons) and one for parameter variability (etas).
    By default the eta hierarchy has the two levels IIV and IOV and the epsilon
    hierarchy has one single level.

    Parameters
    ----------
    rvs : list
        A list of RandomVariable to add. Default is to create an empty RandomVariabels.

    Examples
    --------
    >>> from pharmpy import RandomVariables, RandomVariable, Parameter
    >>> omega = Parameter("OMEGA_CL", 0.1)
    >>> rv = RandomVariable.normal("IIV_CL", "iiv", 0, omega.symbol)
    >>> rvs = RandomVariables([rv])
    """

    def __init__(self, rvs=None):
        if isinstance(rvs, RandomVariables):
            self._rvs = copy.deepcopy(rvs._rvs)
        elif rvs is None:
            self._rvs = []
        else:
            self._rvs = list(rvs)
            names = set()
            for rv in self._rvs:
                if not isinstance(rv, RandomVariable):
                    raise ValueError(f'Can not add variable of type {type(rv)} to RandomVariables')
                if rv.name in names:
                    raise ValueError(
                        f'Names of random variables must be unique. Random Variable "{rv.name}" '
                        'was added more than once to RandomVariables'
                    )
                names.add(rv.name)

        eta_levels = VariabilityHierarchy()
        eta_levels.add_variability_level('IIV', 0, 'ID')
        eta_levels.add_higher_level('IOV', 'OCC')
        epsilon_levels = VariabilityHierarchy()
        epsilon_levels.add_variability_level('RUV', 0, None)
        self._eta_levels = eta_levels
        self._epsilon_levels = epsilon_levels

    def __len__(self):
        return len(self._rvs)

    def __eq__(self, other):
        if len(self) == len(other):
            for s, o in zip(self, other):
                if s != o:
                    return False
            return True
        return False

    def _lookup_rv(self, ind, insert=False):
        if isinstance(ind, sympy.Symbol):
            ind = ind.name
        if isinstance(ind, str):
            for i, rv in enumerate(self._rvs):
                if ind == rv.name:
                    return i, rv
            raise KeyError(f'Could not find {ind} in RandomVariables')
        elif isinstance(ind, RandomVariable):
            try:
                i = self._rvs.index(ind)
            except ValueError:
                raise KeyError(f'Could not find {ind.name} in RandomVariables')
            return i, ind
        if insert:
            # Must allow for inserting after last element.
            return ind, None
        else:
            return ind, self._rvs[ind]

    def __getitem__(self, ind):
        if isinstance(ind, list):
            rvs = []
            for i in ind:
                index, rv = self._lookup_rv(i)
                rvs.append(self[index])
            return RandomVariables(rvs)
        else:
            _, rv = self._lookup_rv(ind)
            return rv

    def __contains__(self, ind):
        try:
            _, _ = self._lookup_rv(ind)
        except KeyError:
            return False
        return True

    def _remove_joint_normal(self, rv):
        # Remove rv from all other rvs
        for other in self:
            if other.name == rv.name:
                continue
            joint_names = other._joint_names
            if joint_names is None or rv.name not in joint_names:
                continue
            joint_index = joint_names.index(rv.name)
            del other._joint_names[joint_index]
            if len(other._joint_names) == 1:
                other._joint_names = None
            other._mean.row_del(joint_index)
            other._variance.row_del(joint_index)
            other._variance.col_del(joint_index)
            other._symengine_variance = symengine.sympify(other._variance)

    def _remove_joint_normal_not_in_self(self):
        # Remove rv from all joint normals not in self
        names = self.names
        for rv in self._rvs:
            if rv._joint_names is not None:
                indices = [i for i, joint_name in enumerate(rv._joint_names) if joint_name in names]
                new_joint = [rv._joint_names[i] for i in indices]
                if len(new_joint) == 1:
                    new_joint = None
                rv._joint_names = new_joint
                means = [rv._mean[i] for i in indices]
                rv._mean = sympy.Matrix(means)
                rv._variance = rv._variance[indices, indices]
                rv._symengine_variance = symengine.sympify(rv._variance)

    def __setitem__(self, ind, value):
        if isinstance(ind, slice):
            if ind.step is None:
                step = 1
            else:
                step = ind.step
            indices = list(range(ind.start, ind.stop, step))
            if len(value) != len(indices):
                raise ValueError('Bad number of rvs to set using slice')
            for i, val in zip(indices, value):
                self[i] = val
            return
        if not isinstance(value, RandomVariable):
            raise ValueError(
                f'Trying to set {type(value)} to RandomVariables. Must be of type RandomVariable.'
            )
        i, rv = self._lookup_rv(ind)
        self.unjoin(rv)
        i, _ = self._lookup_rv(ind)  # Might have moved
        self._rvs[i] = value

    def __delitem__(self, ind):
        i, rv = self._lookup_rv(ind)
        joint_names = rv._joint_names
        if joint_names is not None:
            joint_names = joint_names.copy()
            joint_index = joint_names.index(rv.name)
            for name in joint_names:
                other = self[name]
                del other._joint_names[joint_index]
                if len(other._joint_names) == 1:
                    other._joint_names = None
                other._mean.row_del(joint_index)
                other._variance.row_del(joint_index)
                other._variance.col_del(joint_index)
                other._symengine_variance = symengine.sympify(other._variance)
        del self._rvs[i]

    def __sub__(self, other):
        new = RandomVariables(self._rvs)
        for rv in other:
            if rv in new:
                del new[rv]
        return new

    def insert(self, ind, value):
        if not isinstance(value, RandomVariable):
            raise ValueError(
                f'Trying to insert {type(value)} into RandomVariables. '
                'Must be of type RandomVariable.'
            )
        i, _ = self._lookup_rv(ind, insert=True)
        self._rvs.insert(i, value)

    @property
    def names(self):
        """List of the names of all random variables"""
        return [rv.name for rv in self._rvs]

    @property
    def epsilons(self):
        """Get only the epsilons"""
        return RandomVariables([rv for rv in self._rvs if rv.level in self._epsilon_levels.names])

    @property
    def etas(self):
        """Get only the etas"""
        return RandomVariables([rv for rv in self._rvs if rv.level in self._eta_levels.names])

    @property
    def iiv(self):
        """Get only the iiv etas, i.e. etas with variability level 0"""
        return RandomVariables([rv for rv in self._rvs if rv.level == self._eta_levels.get_name(0)])

    @property
    def iov(self):
        """Get only the iov etas, i.e. etas with variability level 1"""
        return RandomVariables([rv for rv in self._rvs if rv.level == self._eta_levels.get_name(1)])

    @property
    def free_symbols(self):
        """Set of free symbols for all random variables"""
        symbs = set()
        for rv in self._rvs:
            symbs |= rv.free_symbols
        return symbs

    def copy(self):
        """Make copy of RandomVariables"""
        new = RandomVariables()
        for rv in self._rvs:
            new._rvs.append(rv.copy())
        return new

    @property
    def parameter_names(self):
        """List of parameter names for all random variables"""
        params = set()
        for rv in self:
            params |= set(rv.parameter_names)
        return sorted([str(p) for p in params])

    @property
    def variance_parameters(self):
        """List of all parameters representing variance for all random variables"""
        parameters = []
        for rvs, dist in self.distributions():
            if len(rvs) == 1:
                p = dist.std ** 2
                if p not in parameters:
                    parameters.append(dist.std ** 2)
            else:
                for p in dist.sigma.diagonal():
                    if p not in parameters:
                        parameters.append(p)
        return [p.name for p in parameters]

    def get_variance(self, rv):
        """Get variance for a random variable"""
        _, rv = self._lookup_rv(rv)
        if rv._joint_names is None:
            return rv._variance[0]
        else:
            i = rv._joint_names.index(rv.name)
            return rv._variance[i, i]

    def get_covariance(self, rv1, rv2):
        """Get covariance between two random variables"""
        _, rv1 = self._lookup_rv(rv1)
        _, rv2 = self._lookup_rv(rv2)
        if rv1._joint_names is None or rv2.name not in rv1._joint_names:
            return sympy.Integer(0)
        else:
            i1 = rv1._joint_names.index(rv1.name)
            i2 = rv1._joint_names.index(rv2.name)
            return rv1._variance[i1, i2]

    def _rename_rv(self, current, new):
        for rv in self._rvs:
            if rv.name == current:
                rv.name = new
            if rv._joint_names and current in rv._joint_names:
                i = rv._joint_names.index(current)
                rv._joint_names[i] = new

    def subs(self, d):
        """Substitute expressions

        Parameters
        ----------
        d : dict
            Dictionary of from: to pairs for substitution

        Examples
        --------
        >>> import sympy
        >>> from pharmpy import RandomVariables, Parameter
        >>> omega = Parameter("OMEGA_CL", 0.1)
        >>> rv = RandomVariable.normal("IIV_CL", "IIV", 0, omega.symbol)
        >>> rvs = RandomVariables([rv])
        >>> rvs.subs({omega.symbol: sympy.Symbol("OMEGA_NEW")})
        >>> rvs
        IIV_CL ~ 𝒩 (0, OMEGA_NEW)

        """
        s = dict()
        for key, value in d.items():
            key = sympy.sympify(key)
            value = sympy.sympify(value)
            if key.name in self.names:
                self._rename_rv(key.name, value.name)
            else:
                s[key] = value
        for rv in self._rvs:
            rv.subs(s)

    def unjoin(self, inds):
        """Remove all covariances the random variables have with other random variables

        Parameters
        ----------
        inds
            One or multiple indices to unjoin

        Examples
        --------
        >>> from pharmpy import RandomVariables, RandomVariable, Parameter
        >>> omega_cl = Parameter("OMEGA_CL", 0.1)
        >>> omega_v = Parameter("OMEGA_V", 0.1)
        >>> corr_cl_v = Parameter("OMEGA_CL_V", 0.01)
        >>> rv1, rv2 = RandomVariable.joint_normal(["IIV_CL", "IIV_V"], 'IIV', [0, 0],
        ...     [[omega_cl.symbol, corr_cl_v.symbol], [corr_cl_v.symbol, omega_v.symbol]])
        >>> rvs = RandomVariables([rv1, rv2])
        >>> rvs.unjoin('IIV_CL')
        >>> rvs
        IIV_CL ~ 𝒩 (0, OMEGA_CL)
        IIV_V ~ 𝒩 (0, OMEGA_V)

        See Also
        --------
        join
        """
        if not isinstance(inds, list):
            inds = [inds]
        for ind in inds:
            i, rv = self._lookup_rv(ind)
            if rv._joint_names is None:
                return
            index = rv._joint_names.index(rv.name)
            self._remove_joint_normal(rv)
            del self._rvs[i]
            rv._mean = sympy.Matrix([rv._mean[index]])
            rv._variance = sympy.Matrix([rv._variance[index, index]])
            rv._symengine_variance = symengine.sympify(rv._variance)
            rv._joint_names = None
            self._rvs.insert(i - index, rv)

    def join(self, inds, fill=0, name_template=None, param_names=None):
        """Join random variables together into one joint distribution

        Set new covariances (and previous 0 covs) to 'fill'

        Parameters
        ----------
        inds
            Indices of variables to join
        fill : value
            Value to use for new covariances. Default is 0
        name_template : str
            A string template to use for new covariance symbols.
            Using this option will override fill.
        param_names : list
            List of parameter names to be used together with
            name_template.


        Returns
        -------
        A dictionary from newly created covariance parameter names to
        tuple of parameter names. Empty dictionary if no parameter
        symbols were created

        Examples
        --------
        >>> from pharmpy import RandomVariables, RandomVariable, Parameter
        >>> omega_cl = Parameter("OMEGA_CL", 0.1)
        >>> omega_v = Parameter("OMEGA_V", 0.1)
        >>> rv1 = RandomVariable.normal("IIV_CL", 'IIV', 0, omega_cl.symbol)
        >>> rv2 = RandomVariable.normal("IIV_V", 'IIV', 0, omega_v.symbol)
        >>> rvs = RandomVariables([rv1, rv2])
        >>> rvs.join(['IIV_CL', 'IIV_V'])
        {}
        >>> rvs
        ⎡IIV_CL⎤     ⎧⎡0⎤  ⎡OMEGA_CL     0   ⎤⎫
        ⎢      ⎥ ~ 𝒩 ⎪⎢ ⎥, ⎢                 ⎥⎪
        ⎣IIV_V ⎦     ⎩⎣0⎦  ⎣   0      OMEGA_V⎦⎭

        See Also
        --------
        unjoin
        """
        cov_to_params = dict()
        selection = self[inds]
        selection._remove_joint_normal_not_in_self()
        means, M, names, others = selection._calc_covariance_matrix()
        if fill != 0:
            for row, col in itertools.product(range(M.rows), range(M.cols)):
                if M[row, col] == 0:
                    M[row, col] = fill
        elif name_template:
            for row, col in itertools.product(range(M.rows), range(M.cols)):
                if M[row, col] == 0 and row > col:
                    param_1, param_2 = M[row, row], M[col, col]
                    cov_name = name_template.format(param_names[col], param_names[row])
                    cov_to_params[cov_name] = (str(param_1), str(param_2))
                    M[row, col], M[col, row] = symbol(cov_name), symbol(cov_name)

        for i in inds:
            self._remove_joint_normal(self[i])

        new = []
        first = True
        for rv in self._rvs:
            if rv.name in selection:
                if first:
                    new.extend(selection._rvs)
                    first = False
            else:
                new.append(rv)

        new_rvs = RandomVariable.joint_normal(names, 'iiv', means, M)
        for rv, new_rv in zip(selection, new_rvs):
            rv._sympy_rv = new_rv._sympy_rv
            rv._mean = sympy.Matrix(means)
            rv._variance = M.copy()
            rv._symengine_variance = symengine.Matrix(M.rows, M.cols, M)
            rv._joint_names = [rv.name for rv in new_rvs]
        self._rvs = new
        return cov_to_params

    def distributions(self):
        """List with one entry per distribution instead of per random variable.

        Returned is a list of tuples of a list of random variables that are jointly
        distributed and the distribution.

        Example
        -------
        >>> from pharmpy import RandomVariables, RandomVariable, Parameter
        >>> omega_cl = Parameter("OMEGA_CL", 0.1)
        >>> omega_v = Parameter("OMEGA_V", 0.1)
        >>> omega_ka = Parameter("OMEGA_KA", 0.1)
        >>> corr_cl_v = Parameter("OMEGA_CL_V", 0.01)
        >>> rv1, rv2 = RandomVariable.joint_normal(["IIV_CL", "IIV_V"], 'IIV', [0, 0],
        ...     [[omega_cl.symbol, corr_cl_v.symbol], [corr_cl_v.symbol, omega_v.symbol]])
        >>> rv3 = RandomVariable.normal("IIV_KA", 'IIV', 0, omega_ka.symbol)
        >>> rvs = RandomVariables([rv1, rv2, rv3])
        >>> dists = rvs.distributions()

        """
        distributions = []
        i = 0
        while i < len(self):
            rv = self[i]
            symrv = rv.sympy_rv
            n = 1 if rv._joint_names is None else len(rv._joint_names)
            dist = symrv.pspace.distribution
            if isinstance(dist, stats.crv_types.NormalDistribution):
                i += 1
                distributions.append(([rv], dist))
            else:  # Joint Normal
                rvs = [self[k] for k in range(i, i + n)]
                i += n
                distributions.append((rvs, dist))
        return distributions

    def nearest_valid_parameters(self, parameter_values):
        """Force parameter values into being valid

        As small changes as possible

        returns a dict with the valid parameter values
        """
        nearest = parameter_values.copy()
        for rvs, dist in self.distributions():
            if len(rvs) > 1:
                symb_sigma = rvs[0]._variance
                sigma = symb_sigma.subs(dict(parameter_values))
                A = np.array(sigma).astype(np.float64)
                B = pharmpy.math.nearest_posdef(A)
                if B is not A:
                    for row in range(len(A)):
                        for col in range(row + 1):
                            nearest[symb_sigma[row, col].name] = B[row, col]
        return nearest

    def validate_parameters(self, parameter_values):
        """Validate a dict or Series of parameter values

        Currently checks that all covariance matrices are posdef
        use_cache for using symengine cached matrices
        """
        for rvs, dist in self.distributions():
            if len(rvs) > 1:
                sigma = rvs[0]._symengine_variance
                replacement = {}
                for param in dict(parameter_values):
                    replacement[symengine.Symbol(param)] = parameter_values[param]
                sigma = sigma.subs(replacement)
                if not sigma.free_symbols:  # Cannot validate since missing params
                    a = np.array(sigma).astype(np.float64)
                    if not pharmpy.math.is_posdef(a):
                        return False
                else:
                    raise TypeError("Cannot validate parameters since all are not numeric")
        return True

    def sample(self, expr, parameters=None, samples=1, rng=None):
        """Sample from the distribution of expr

        parameters in the distriutions will first be replaced"""
        rng = _create_rng(rng)
        if not parameters:
            parameters = dict()
        expr = sympy.sympify(expr).subs(parameters)
        symbols = expr.free_symbols
        expr_names = [symb.name for symb in symbols]
        i = 0
        sampling_rvs = []
        for rvs, dist in self.distributions():
            names = [rv.name for rv in rvs]
            if set(names) & set(expr_names):
                new_name = f'__J{i}'
                if len(rvs) > 1:
                    mu = dist.mu.subs(parameters)
                    sigma = dist.sigma.subs(parameters)
                else:
                    mu = dist.mean.subs(parameters)
                    sigma = dist.std.subs(parameters)
                new_rv = sympy.stats.Normal(new_name, mu, sigma)
                sampling_rvs.append((names, new_rv))
        if sampling_rvs:
            # FIXME: Unnecessary to go via DataFrame
            df = pd.DataFrame(index=range(samples))
            for names, new_rv in sampling_rvs:
                with warnings.catch_warnings():
                    warnings.filterwarnings('ignore')
                    if sympy.__version__ == '1.8':
                        cursample = next(
                            sympy.stats.sample(new_rv, library='numpy', size=samples, seed=rng)
                        )
                    else:
                        cursample = sympy.stats.sample(
                            new_rv, library='numpy', size=samples, seed=rng
                        )
                    if len(names) > 1:
                        df[names] = cursample
                    else:
                        df[names[0]] = cursample
            ordered_symbols = list(symbols)
            input_list = [df[symb.name].values for symb in ordered_symbols]
            fn = sympy.lambdify(ordered_symbols, expr, 'numpy')
            a = fn(*input_list)
        else:
            a = np.full(samples, float(expr.evalf()))
        return a

    def _calc_covariance_matrix(self):
        non_altered = []
        means = []
        blocks = []
        names = []
        for rvs, dist in self.distributions():
            names.extend([rv.name for rv in rvs])
            if isinstance(dist, stats.crv_types.NormalDistribution):
                means.append(dist.mean)
                blocks.append(sympy.Matrix([dist.std ** 2]))
            elif isinstance(dist, stats.joint_rv_types.MultivariateNormalDistribution):
                means.extend(dist.mu)
                blocks.append(dist.sigma)
            else:
                non_altered.extend(rvs)
        if names:
            M = sympy.BlockDiagMatrix(*blocks)
            M = sympy.Matrix(M)
        else:
            M = sympy.Matrix()
        return means, M, names, non_altered

    @property
    def covariance_matrix(self):
        """Covariance matrix of all random variables"""
        _, M, _, others = self._calc_covariance_matrix()
        if others:
            raise ValueError('Only normal distributions are supported')
        return M

    def __repr__(self):
        strings = []
        for rvs, dist in self.distributions():
            strings.append(repr(rvs[0]))
        return '\n'.join(strings)

    def _repr_latex_(self):
        lines = []
        for rvs, dist in self.distributions():
            rv = rvs[0]
            latex = rv._latex_string(aligned=True)
            lines.append(latex)
        return '\\begin{align*}\n' + r' \\ '.join(lines) + '\\end{align*}'

    def parameters_sdcorr(self, values):
        """Convert parameter values to sd/corr form

        All parameter values will be converted to sd/corr assuming
        they are given in var/cov form. Only parameters for normal
        distributions will be affected.

        Parameters
        ----------
        values : dict
            Dict of parameter names to values
        """
        newdict = values.copy()
        for rvs, dist in self.distributions():
            if len(rvs) > 1:
                sigma_sym = dist.sigma
                sigma = np.array(sigma_sym.subs(values)).astype(np.float64)
                corr = pharmpy.math.cov2corr(sigma)
                for i in range(sigma_sym.rows):
                    for j in range(sigma_sym.cols):
                        name = sigma_sym[i, j].name
                        if i != j:
                            newdict[name] = corr[i, j]
                        else:
                            newdict[name] = np.sqrt(sigma[i, j])
            else:
                name = (dist.std ** 2).name
                if name in newdict:
                    newdict[name] = dist.std.subs(values)
        return newdict



# All functions used for comparing a set of candidate models and
# ranking them take the same base arguments
# base - the base model
# candidates - a list of candidate models
# returns a list which is a subset of the candidates with the best ranked at the beginning
# models can be removed from candidates if they don't pass some criteria


def ofv(base, candidates, cutoff=3.84):
    base_ofv = base.modelfit_results.ofv
    filtered = [model for model in candidates if base_ofv - model.modelfit_results.ofv >= cutoff]
    srtd = sorted(filtered, key=lambda model: model.modelfit_results.ofv)
    return srtd


def aic(base, candidates, cutoff=3.84):
    base_aic = base.modelfit_results.aic
    filtered = [model for model in candidates if base_aic - model.modelfit_results.aic >= cutoff]
    srtd = sorted(filtered, key=lambda model: model.modelfit_results.aic)
    return srtd


def bic(base, candidates, cutoff=3.84):
    base_bic = base.modelfit_results.bic
    filtered = [model for model in candidates if base_bic - model.modelfit_results.bic >= cutoff]
    srtd = sorted(filtered, key=lambda model: model.modelfit_results.bic)
    return srtd



import itertools

import numpy as np

import pharmpy.math


def calculate_parcov_inits(model, ncovs):
    """Get dict of new updated inits for parcov block

    model already has a FREM block
    Initial estimates for the parcov block is calculated given correlations of individual etas
    """
    rvs, dist = model.random_variables.iiv.distributions()[-1]
    rvs = [rv.name for rv in rvs]
    ie = model.modelfit_results.individual_estimates
    eta_corr = ie[rvs].corr()
    eta_corr.fillna(value=1.0, inplace=True)  # Identical etas will get NaN as both diag and corr

    sigma = dist.sigma
    inits = sigma.subs(model.parameters.inits)
    inits = np.array(inits).astype(np.float64)
    sd = np.sqrt(inits.diagonal())
    npars = len(sd) - ncovs

    cov = pharmpy.math.corr2cov(eta_corr.to_numpy(), sd)
    cov[cov == 0] = 0.0001
    cov = pharmpy.math.nearest_posdef(cov)

    parcov_inits = cov[npars:, :npars]
    parcov_symb = sigma[npars:, :npars]

    param_inits = {
        parcov_symb[i, j].name: parcov_inits[i, j]
        for i, j in itertools.product(range(ncovs), range(npars))
    }
    return param_inits


def create_model3b(model1b, model3, ncovs):
    """Create model 3b from model 3

    * Update parcov omega block
    * Set FIX pattern back from model1b
    * Use initial etas from model3
    """
    parameters = model3.parameters.copy()

    parcov_inits = calculate_parcov_inits(model3, ncovs)
    parameters.inits = parcov_inits
    parameters.inits = model3.random_variables.nearest_valid_parameters(parameters.inits)
    parameters.fix = model1b.parameters.fix

    model3b = model3.copy()
    model3b.parameters = parameters
    model3b.name = 'model_3b'
    model3b.update_individual_estimates(model3)
    return model3b



import os
import re
import shutil
import warnings
from pathlib import Path
from urllib.request import urlopen

from bs4 import BeautifulSoup
from csscompressor import compress
from sphinx.application import Sphinx

from pharmpy.utils import TemporaryDirectory, TemporaryDirectoryChanger


def generate_report(rst_path, results_path):
    """Generate report from rst and results json"""
    results_path = Path(results_path)
    with TemporaryDirectory() as tmpdirname:
        tmp_path = Path(tmpdirname)
        source_path = tmp_path / 'source'
        source_path.mkdir()
        shutil.copy(rst_path, source_path / 'results.rst')
        shutil.copy(results_path / 'results.json', source_path)

        conf_path = Path(__file__).parent

        # Change directory for results.json to be found
        with TemporaryDirectoryChanger(source_path):
            with open(os.devnull, 'w') as devnull:
                with warnings.catch_warnings():
                    # Don't display deprecation warnings.
                    # See https://github.com/pharmpy/pharmpy/issues/20
                    warnings.filterwarnings("ignore", message="The app.add_stylesheet")
                    warnings.filterwarnings("ignore", message="The app.add_javascript")
                    app = Sphinx(
                        str(source_path),
                        str(conf_path),
                        str(tmp_path),
                        str(tmp_path),
                        "singlehtml",
                        status=devnull,
                        warning=devnull,
                    )
                    app.build()

        # Write missing altair css
        with open(tmp_path / '_static' / 'altair-plot.css', 'w') as dh:
            dh.write(
                """.vega-actions a {
    margin-right: 12px;
    color: #757575;
    font-weight: normal;
    font-size: 13px;
}

.vega-embed {
    margin-bottom: 20px;
    margin-top: 20px;
    width: 100%;
}
"""
            )
        report_path = tmp_path / 'results.html'
        embed_css_and_js(tmp_path / 'results.html', report_path)
        shutil.copy(report_path, results_path)


def embed_css_and_js(html, target):
    """Embed all external css and javascript into an html"""
    with open(html, 'r', encoding='utf-8') as sh:
        soup = BeautifulSoup(sh, features='lxml')

    scripts = soup.findAll("script", attrs={"src": True})

    for script in scripts:
        source = script.attrs['src']
        if source.startswith('http'):
            infile = urlopen(source)
            content = infile.read().decode('utf-8')
            infile.close()
        else:
            path = html.parent / source
            with open(path, 'r') as sh:
                content = sh.read()

        # Minification with jsmin didn't work
        tag = soup.new_tag('script')
        tag['type'] = 'text/javascript'
        tag.append(content)
        script.replace_with(tag)

    stylesheets = soup.findAll("link", attrs={"rel": "stylesheet"})

    for stylesheet in stylesheets:
        stylesheet_src = stylesheet.attrs['href']
        tag = soup.new_tag("style")
        tag['type'] = 'text/css'
        path = html.parent / stylesheet_src
        with open(path, 'r') as sh:
            content = sh.read()
        if '@import' in content:
            import_files = re.findall(r'@import\s+url\("([A-Za-z0-9.]+)"\)', content)
            for name in import_files:
                with open(path.parent / name, 'r') as import_file:
                    import_content = import_file.read()
                content = re.sub(r'@import\s+url\("' + name + r'"\);', import_content, content)
        minified_content = compress(content)
        tag.append(minified_content)
        stylesheet.replace_with(tag)

    with open(target, 'w', encoding='utf-8') as dh:
        dh.write(str(soup))



import copy
import itertools
import re
import warnings
from pathlib import Path

import numpy as np
import sympy

from pharmpy import data
from pharmpy.plugins.nonmem.records import code_record
from pharmpy.random_variables import RandomVariables
from pharmpy.statements import (
    Assignment,
    Bolus,
    CompartmentalSystem,
    ExplicitODESystem,
    Infusion,
    ModelStatements,
    ODESystem,
)
from pharmpy.symbols import symbol

from .records.factory import create_record


def update_parameters(model, old, new):
    new_names = {p.name for p in new}
    old_names = {p.name for p in old}
    removed = old_names - new_names
    if removed:
        remove_records = []
        next_theta = 1
        for theta_record in model.control_stream.get_records('THETA'):
            current_names = theta_record.name_map.keys()
            if removed >= current_names:
                remove_records.append(theta_record)
            elif not removed.isdisjoint(current_names):
                # one or more in the record
                theta_record.remove(removed & current_names)
                theta_record.renumber(next_theta)
                next_theta += len(theta_record)
            else:
                # keep all
                theta_record.renumber(next_theta)
                next_theta += len(theta_record)
        for sigma_record in model.control_stream.get_records('SIGMA'):
            current_names = sigma_record.name_map.keys()
            if removed >= current_names:
                remove_records.append(sigma_record)
        model.control_stream.remove_records(remove_records)

    for p in new:
        name = p.name
        if name not in old and name not in model.random_variables.parameter_names:
            # This is a new theta
            theta_number = get_next_theta(model)
            record = create_theta_record(model, p)
            if re.match(r'THETA\(\d+\)', name):
                p.name = f'THETA({theta_number})'
            record.add_nonmem_name(p.name, theta_number)

    next_theta = 1
    for theta_record in model.control_stream.get_records('THETA'):
        theta_record.update(new, next_theta)
        next_theta += len(theta_record)
    next_omega = 1
    previous_size = None
    for omega_record in model.control_stream.get_records('OMEGA'):
        next_omega, previous_size = omega_record.update(new, next_omega, previous_size)
    next_sigma = 1
    previous_size = None
    for sigma_record in model.control_stream.get_records('SIGMA'):
        next_sigma, previous_size = sigma_record.update(new, next_sigma, previous_size)


def update_random_variables(model, old, new):
    from pharmpy.plugins.nonmem.records.code_record import diff

    if not hasattr(model, '_parameters'):
        model.parameters

    rec_dict = dict()
    comment_dict = dict()

    for omega_record in model.control_stream.get_records(
        'OMEGA'
    ) + model.control_stream.get_records('SIGMA'):
        comment_dict = {**comment_dict, **omega_record.comment_map}
        current_names = list(omega_record.eta_map.keys())
        for name in current_names:
            rec_dict[name] = omega_record

    rvs_diff_eta = diff(old.etas.distributions(), new.etas.distributions())
    rvs_diff_eps = diff(old.epsilons.distributions(), new.epsilons.distributions())

    update_random_variable_records(model, rvs_diff_eta, rec_dict, comment_dict)
    update_random_variable_records(model, rvs_diff_eps, rec_dict, comment_dict)

    next_eta = 1
    for omega_record in model.control_stream.get_records('OMEGA'):
        omega_record.renumber(next_eta)
        next_eta += len(omega_record)

    next_eps = 1
    for sigma_record in model.control_stream.get_records('SIGMA'):
        sigma_record.renumber(next_eps)
        next_eps += len(sigma_record)


def update_random_variable_records(model, rvs_diff, rec_dict, comment_dict):
    removed = []
    eta_number = 1
    number_of_records = 0

    rvs_diff = list(rvs_diff)

    rvs_removed = [RandomVariables(rvs).names for (op, (rvs, _)) in rvs_diff if op == '-']
    rvs_removed = [rv for sublist in rvs_removed for rv in sublist]

    for i, (op, (rvs, _)) in enumerate(rvs_diff):
        if op == '+':
            if len(rvs) == 1:
                create_omega_single(model, rvs[0], eta_number, number_of_records, comment_dict)
            else:
                create_omega_block(model, rvs, eta_number, number_of_records, comment_dict)
            eta_number += len(rvs)
            number_of_records += 1
        elif op == '-':
            rvs_rec = list({rec_dict[rv.name] for rv in rvs})
            recs_to_remove = [rec for rec in rvs_rec if rec not in removed]
            if recs_to_remove:
                model.control_stream.remove_records(recs_to_remove)
                removed += [rec for rec in recs_to_remove]
        else:
            diag_rvs = get_diagonal(rvs[0], rec_dict)
            # Account for etas in diagonal
            if len(rvs) == 1 and diag_rvs and any(rv in rvs_removed for rv in diag_rvs):
                create_omega_single(model, rvs[0], eta_number, number_of_records, comment_dict)
            eta_number += len(rvs)
            number_of_records += 1


def get_diagonal(rv, rec_dict):
    rv_rec = rec_dict[rv.name]
    etas_from_same_record = [eta for eta, rec in rec_dict.items() if rec == rv_rec]
    if len(etas_from_same_record) > 1:
        return etas_from_same_record
    return None


def get_next_theta(model):
    """Find the next available theta number"""
    next_theta = 1

    for theta_record in model.control_stream.get_records('THETA'):
        thetas = theta_record.parameters(next_theta)
        next_theta += len(thetas)

    return next_theta


def create_theta_record(model, param):
    param_str = '$THETA  '

    if param.upper < 1000000:
        if param.lower <= -1000000:
            param_str += f'(-INF,{param.init},{param.upper})'
        else:
            param_str += f'({param.lower},{param.init},{param.upper})'
    else:
        if param.lower <= -1000000:
            param_str += f'{param.init}'
        else:
            param_str += f'({param.lower},{param.init})'
    if param.fix:
        param_str += ' FIX'
    param_str += '\n'
    record = model.control_stream.insert_record(param_str)
    return record


def create_omega_single(model, rv, eta_number, record_number, comment_dict):
    rvs, pset = model.random_variables, model.parameters

    if rv.level == 'RUV':
        record_type = 'SIGMA'
    else:
        record_type = 'OMEGA'

    variance_param = pset[rv.parameter_names[0]]

    if rv.level == 'IOV':
        param_str = f'${record_type}  BLOCK(1)'
        first_iov = [iov for iov in rvs.iov if iov.parameter_names == rv.parameter_names][0]
        if rv == first_iov:
            param_str += f'\n{variance_param.init}'
        else:
            param_str += ' SAME'
    else:
        param_str = f'${record_type}  {variance_param.init}'

    if not re.match(r'(OMEGA|SIGMA)\(\d+,\d+\)', variance_param.name):
        param_str += f' ; {variance_param.name}'
    elif comment_dict and variance_param.name in comment_dict.keys():
        param_str += f' ; {comment_dict[variance_param.name]}'

    record = insert_omega_record(model, f'{param_str}\n', record_number, record_type)

    record.comment_map = comment_dict
    record.eta_map = {rv.name: eta_number}
    record.name_map = {variance_param.name: (eta_number, eta_number)}


def create_omega_block(model, rvs, eta_number, record_number, comment_dict):
    rvs = RandomVariables(rvs)
    param_str = f'$OMEGA BLOCK({rvs.covariance_matrix.shape[0]})\n'

    for row in range(rvs.covariance_matrix.shape[0]):
        for col in range(row + 1):
            elem = rvs.covariance_matrix.row(row).col(col)
            name = str(elem[0])
            omega = model.parameters[name]
            param_str += f'{omega.init}'

            if not re.match(r'OMEGA\(\d+,\d+\)', omega.name):
                param_str += f'\t; {omega.name}'
            elif comment_dict and omega.name in comment_dict:
                param_str += f'\t; {comment_dict[omega.name]}'

            param_str += '\n'

        param_str = f'{param_str.rstrip()}\n'

    eta_map, name_variance = dict(), dict()

    for rv in rvs:
        variance_param = rvs.get_variance(rv)
        eta_map[rv.name] = eta_number
        name_variance[variance_param.name] = (eta_number, eta_number)
        eta_number += 1

    rv_combinations = [
        (rv1.name, rv2.name) for idx, rv1 in enumerate(rvs) for rv2 in rvs[idx + 1 :]
    ]
    name_covariance = {
        rvs.get_covariance(rv1, rv2).name: (eta_map[rv2], eta_map[rv1])
        for rv1, rv2 in rv_combinations
    }

    record = insert_omega_record(model, param_str, record_number, 'OMEGA')

    record.comment_map = comment_dict
    record.eta_map = eta_map
    record.name_map = {**name_variance, **name_covariance}


def insert_omega_record(model, param_str, record_number, record_type):
    records = model.control_stream.records
    tprecs = model.control_stream.get_records(record_type)
    if tprecs:
        index = records.index(tprecs[0])
        record = model.control_stream.insert_record(param_str, index + record_number)
    else:
        record = model.control_stream.insert_record(param_str)
    return record


def update_ode_system(model, old, new):
    """Update ODE system

    Handle changes from CompartmentSystem to ExplicitODESystem
    """
    if type(old) == CompartmentalSystem and type(new) == ExplicitODESystem:
        to_des(model, new)
    elif type(old) == ExplicitODESystem and type(new) == CompartmentalSystem:
        # Stay with $DES for now
        update_des(model, old, new)
    elif type(old) == CompartmentalSystem and type(new) == CompartmentalSystem:
        # subs = model.control_stream.get_records('SUBROUTINES')[0]
        # old_trans = subs.get_option_startswith('TRANS')
        # conv_advan, new_advan, new_trans = change_advan(model)
        update_lag_time(model, old, new)
        if isinstance(new.dosing_compartment.dose, Bolus) and 'RATE' in model.dataset.columns:
            df = model.dataset
            df.drop(columns=['RATE'], inplace=True)
            model.dataset = df

        advan, trans = new_advan_trans(model)
        pk_param_conversion(model, advan=advan, trans=trans)
        add_needed_pk_parameters(model, advan, trans)
        update_subroutines_record(model, advan, trans)
        update_model_record(model, advan)

        statements = model.statements
        if isinstance(new.dosing_compartment.dose, Infusion) and not statements.find_assignment(
            'D1'
        ):
            # Handle direct moving of Infusion dose
            statements.subs({'D2': 'D1'})

        if isinstance(new.dosing_compartment.dose, Infusion) and isinstance(
            old.dosing_compartment.dose, Bolus
        ):
            dose = new.dosing_compartment.dose
            if dose.rate is None:
                # FIXME: Not always D1 here!
                ass = Assignment('D1', dose.duration)
                dose.duration = ass.symbol
            else:
                raise NotImplementedError("First order infusion rate is not yet supported")
            statements = model.statements
            statements.insert_before_odes(ass)
            df = model.dataset
            rate = np.where(df['AMT'] == 0, 0, -2)
            df['RATE'] = rate
            # FIXME: Adding at end for now. Update $INPUT cannot yet handle adding in middle
            # df.insert(list(df.columns).index('AMT') + 1, 'RATE', rate)
            model.dataset = df

    force_des(model, new)


def update_des(model, old, new):
    """Update where both old and new should be explicit ODE systems"""
    pass


def force_des(model, odes):
    """Switch to $DES if necessary"""
    if isinstance(odes, ExplicitODESystem):
        return

    amounts = {sympy.Function(amt.name)(symbol('t')) for amt in odes.amounts}
    if odes.atoms(sympy.Function) & amounts:
        explicit_odes(model)
        new = model.statements.ode_system
        to_des(model, new)


def explicit_odes(model):
    """Convert model from compartmental system to explicit ODE system
    or do nothing if it already has an explicit ODE system

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to same model
    """
    statements = model.statements
    odes = statements.ode_system
    if isinstance(odes, CompartmentalSystem):
        new = odes.to_explicit_system()
        statements[model.statements.index(odes)] = new
        model.statements = statements
        new.solver = odes.solver
    return model


def to_des(model, new):
    subs = model.control_stream.get_records('SUBROUTINES')[0]
    subs.remove_option_startswith('TRANS')
    subs.remove_option_startswith('ADVAN')
    if new.solver:
        if new.solver == 'LSODA':
            advan = 'ADVAN13'
        elif new.solver == 'CVODES':
            advan = 'ADVAN14'
        elif new.solver == 'DGEAR':
            advan = 'ADVAN8'
        elif new.solver == 'DVERK':
            advan = 'ADVAN6'
        elif new.solver == 'IDA':
            advan = 'ADVAN15'
        elif new.solver == 'LSODA':
            advan = 'ADVAN13'
        elif new.solver == 'LSODI':
            advan = 'ADVAN9'
        subs.append_option(advan)
    else:
        subs.append_option('ADVAN6')
    subs.append_option('TOL', 3)
    des = model.control_stream.insert_record('$DES\nDUMMY=0')
    des.from_odes(new)
    model.control_stream.remove_records(model.control_stream.get_records('MODEL'))
    mod = model.control_stream.insert_record('$MODEL\n')
    for eq, ic in zip(new.odes[:-1], list(new.ics.keys())[:-1]):
        name = eq.lhs.args[0].name[2:]
        if new.ics[ic] != 0:
            dose = True
        else:
            dose = False
        mod.add_compartment(name, dosing=dose)


def update_statements(model, old, new, trans):
    trans['NaN'] = int(data.conf.na_rep)
    main_statements = ModelStatements()
    error_statements = ModelStatements()

    new_odes = new.ode_system
    if new_odes is not None:
        old_odes = old.ode_system
        if new_odes != old_odes:
            colnames, _, drop, _ = model._column_info()
            col_dropped = {key: value for key, value in zip(colnames, drop)}
            if 'CMT' in col_dropped.keys() and not col_dropped['CMT']:
                warnings.warn(
                    'Compartment structure has been updated, CMT-column '
                    'in dataset might not be relevant anymore. Check '
                    'CMT-column or drop column'
                )
            update_ode_system(model, old_odes, new_odes)

    after_odes = False
    for s in new:
        if isinstance(s, ODESystem):
            after_odes = True
        elif after_odes:
            error_statements.append(copy.deepcopy(s))
        else:
            main_statements.append(copy.deepcopy(s))

    main_statements.subs(trans)

    rec = model.get_pred_pk_record()
    rec.rvs, rec.trans = model.random_variables, trans
    rec.statements = main_statements

    error = model._get_error_record()
    if error:
        if len(error_statements) > 0:
            error_statements.pop(0)  # Remove the link statement
        error.rvs, error.trans = model.random_variables, trans
        error_statements.subs(trans)
        error.statements = error_statements
        error.is_updated = True

    rec.is_updated = True


def update_lag_time(model, old, new):
    new_dosing = new.dosing_compartment
    new_lag_time = new_dosing.lag_time
    old_lag_time = old.dosing_compartment.lag_time
    if new_lag_time != old_lag_time and new_lag_time != 0:
        ass = Assignment('ALAG1', new_lag_time)
        model.statements.insert_before_odes(ass)
        new_dosing.lag_time = ass.symbol


def new_compartmental_map(cs, oldmap):
    """Create compartmental map for updated model
    cs - new compartmental system
    old - old compartmental map

    Can handle compartments from dosing to central, peripherals and output
    """
    comp = cs.dosing_compartment
    central = cs.central_compartment
    i = 1
    compmap = dict()
    while True:
        compmap[comp.name] = i
        i += 1
        if comp is central:
            break
        comp, _ = cs.get_compartment_outflows(comp)[0]

    peripherals = cs.peripheral_compartments
    for p in peripherals:
        compmap[p.name] = i
        i += 1

    diff = len(cs) - len(oldmap)
    for name in cs.names:
        if name not in compmap.keys():
            compmap[name] = oldmap[name] + diff
    return compmap


def create_compartment_remap(oldmap, newmap):
    """Creates a map from old compartment number to new compartment number

    For all compartments where remapping is needed
    Assume that compartments with same name in new and old are the same compartments
    """
    remap = dict()
    for name, number in oldmap.items():
        if name in newmap:
            remap[number] = newmap[name]
    return remap


def pk_param_conversion(model, advan, trans):
    """Conversion map for pk parameters for removed or added compartment"""
    subs = model.control_stream.get_records('SUBROUTINES')[0]
    from_advan = subs.get_option_startswith('ADVAN')
    statements = model.statements
    cs = statements.ode_system
    oldmap = model._compartment_map
    newmap = new_compartmental_map(cs, oldmap)
    remap = create_compartment_remap(oldmap, newmap)
    d = dict()
    for old, new in remap.items():
        d[symbol(f'S{old}')] = symbol(f'S{new}')
        d[symbol(f'F{old}')] = symbol(f'F{new}')
        d[symbol(f'R{old}')] = symbol(f'R{new}')
        d[symbol(f'D{old}')] = symbol(f'D{new}')
        d[symbol(f'ALAG{old}')] = symbol(f'ALAG{new}')
        d[symbol(f'A({old})')] = symbol(f'A({new})')
    if from_advan == 'ADVAN5' or from_advan == 'ADVAN7':
        reverse_map = {v: k for k, v in newmap.items()}
        for i, j in itertools.product(range(1, len(oldmap)), range(0, len(oldmap))):
            if i != j and (i in remap and (j in remap or j == 0)):
                if i in remap:
                    to_i = remap[i]
                else:
                    to_i = i
                if j in remap:
                    to_j = remap[j]
                else:
                    to_j = j
                outind = to_j if to_j != 0 else len(cs)
                from_comp = cs.find_compartment(reverse_map[to_i])
                to_comp = cs.find_compartment(reverse_map[outind])
                if cs.get_flow(from_comp, to_comp) is not None:
                    d[symbol(f'K{i}{j}')] = symbol(f'K{to_i}{to_j}')
                    d[symbol(f'K{i}T{j}')] = symbol(f'K{to_i}T{to_j}')
        if advan == 'ADVAN3':
            n = len(oldmap)
            for i in range(1, n):
                d[symbol(f'K{i}0')] = symbol('K')
                d[symbol(f'K{i}T0')] = symbol('K')
                d[symbol(f'K{i}{n}')] = symbol('K')
                d[symbol(f'K{i}T{n}')] = symbol('K')
    elif from_advan == 'ADVAN1':
        if advan == 'ADVAN3' or advan == 'ADVAN11':
            d[symbol('V')] = symbol('V1')
        elif advan == 'ADVAN4' or advan == 'ADVAN12':
            d[symbol('V')] = symbol('V2')
    elif from_advan == 'ADVAN2':
        if advan == 'ADVAN3' and trans != 'TRANS1':
            d[symbol('V')] = symbol('V1')
        elif advan == 'ADVAN4' and trans != 'TRANS1':
            d[symbol('V')] = symbol('V2')
    elif from_advan == 'ADVAN3':
        if advan == 'ADVAN1':
            if trans == 'TRANS2':
                d[symbol('V1')] = symbol('V')
        elif advan == 'ADVAN4':
            if trans == 'TRANS4':
                d[symbol('V1')] = symbol('V2')
                d[symbol('V2')] = symbol('V3')
            elif trans == 'TRANS6':
                d[symbol('K21')] = symbol('K32')
            else:  # TRANS1
                d[symbol('K12')] = symbol('K23')
                d[symbol('K21')] = symbol('K32')
        elif advan == 'ADVAN11':
            if trans == 'TRANS4':
                d.update({symbol('Q'): symbol('Q2')})
    elif from_advan == 'ADVAN4':
        if advan == 'ADVAN2':
            if trans == 'TRANS2':
                d[symbol('V2')] = symbol('V')
        if advan == 'ADVAN3':
            if trans == 'TRANS4':
                d.update({symbol('V2'): symbol('V1'), symbol('V3'): symbol('V2')})
            elif trans == 'TRANS6':
                d.update({symbol('K32'): symbol('K21')})
            else:  # TRANS1
                d.update({symbol('K23'): symbol('K12'), symbol('K32'): symbol('K21')})
        elif advan == 'ADVAN12':
            if trans == 'TRANS4':
                d.update({symbol('Q'): symbol('Q3')})
    elif from_advan == 'ADVAN11':
        if advan == 'ADVAN1':
            if trans == 'TRANS2':
                d[symbol('V1')] = symbol('V')
        elif advan == 'ADVAN3':
            if trans == 'TRANS4':
                d[symbol('Q2')] = symbol('Q')
        elif advan == 'ADVAN12':
            if trans == 'TRANS4':
                d.update(
                    {
                        symbol('V1'): symbol('V2'),
                        symbol('Q2'): symbol('Q3'),
                        symbol('V2'): symbol('V3'),
                        symbol('Q3'): symbol('Q4'),
                        symbol('V3'): symbol('V4'),
                    }
                )
            elif trans == 'TRANS6':
                d.update({symbol('K31'): symbol('K42'), symbol('K21'): symbol('K32')})
            else:  # TRANS1
                d.update(
                    {
                        symbol('K12'): symbol('K23'),
                        symbol('K21'): symbol('K32'),
                        symbol('K13'): symbol('K24'),
                        symbol('K31'): symbol('K42'),
                    }
                )
    elif from_advan == 'ADVAN12':
        if advan == 'ADVAN2':
            if trans == 'TRANS2':
                d[symbol('V2')] = symbol('V')
        elif advan == 'ADVAN4':
            if trans == 'TRANS4':
                d[symbol('Q3')] = symbol('Q')
        elif advan == 'ADVAN11':
            if trans == 'TRANS4':
                d.update(
                    {
                        symbol('V2'): symbol('V1'),
                        symbol('Q3'): symbol('Q2'),
                        symbol('V3'): symbol('V2'),
                        symbol('Q4'): symbol('Q3'),
                        symbol('V4'): symbol('V3'),
                    }
                )
            elif trans == 'TRANS6':
                d.update({symbol('K42'): symbol('K31'), symbol('K32'): symbol('K21')})
            else:  # TRANS1
                d.update(
                    {
                        symbol('K23'): symbol('K12'),
                        symbol('K32'): symbol('K21'),
                        symbol('K24'): symbol('K13'),
                        symbol('K42'): symbol('K31'),
                    }
                )
    if advan == 'ADVAN5' or advan == 'ADVAN7' and from_advan not in ('ADVAN5', 'ADVAN7'):
        n = len(newmap)
        d[symbol('K')] = symbol(f'K{n-1}0')
    statements.subs(d)


def new_advan_trans(model):
    """Decide which new advan and trans to be used"""
    subs = model.control_stream.get_records('SUBROUTINES')[0]
    oldtrans = subs.get_option_startswith('TRANS')
    statements = model.statements
    odes = model.statements.ode_system
    if len(odes) > 5 or odes.get_n_connected(odes.central_compartment) != len(odes) - 1:
        advan = 'ADVAN5'
    elif len(odes) == 2:
        advan = 'ADVAN1'
    elif len(odes) == 3 and odes.find_depot(statements):
        advan = 'ADVAN2'
    elif len(odes) == 3:
        advan = 'ADVAN3'
    elif len(odes) == 4 and odes.find_depot(statements):
        advan = 'ADVAN4'
    elif len(odes) == 4:
        advan = 'ADVAN11'
    else:  # len(odes) == 5
        advan = 'ADVAN12'

    if oldtrans == 'TRANS1':
        trans = oldtrans
    elif oldtrans == 'TRANS2':
        if advan in ['ADVAN1', 'ADVAN2']:
            trans = oldtrans
        elif advan in ['ADVAN3', 'ADVAN4', 'ADVAN11', 'ADVAN12']:
            trans = 'TRANS4'
        else:
            trans = 'TRANS1'
    elif oldtrans == 'TRANS3':
        if advan in ['ADVAN3', 'ADVAN4']:
            trans = oldtrans
        elif advan in ['ADVAN11', 'ADVAN12']:
            trans = 'TRANS4'
        elif advan in ['ADVAN1', 'ADVAN2']:
            trans = 'TRANS2'
        else:
            trans = 'TRANS1'
    elif oldtrans == 'TRANS4':
        if advan in ['ADVAN3', 'ADVAN4', 'ADVAN11', 'ADVAN12']:
            trans = oldtrans
        elif advan in ['ADVAN1', 'ADVAN2']:
            trans = 'TRANS2'
        else:
            trans = 'TRANS1'
    else:
        trans = 'TRANS1'

    return advan, trans


def update_subroutines_record(model, advan, trans):
    """Update $SUBROUTINES with new advan and trans"""
    subs = model.control_stream.get_records('SUBROUTINES')[0]
    oldadvan = subs.get_option_startswith('ADVAN')
    oldtrans = subs.get_option_startswith('TRANS')

    if advan != oldadvan:
        subs.replace_option(oldadvan, advan)
    if trans != oldtrans:
        subs.replace_option(oldtrans, trans)


def update_model_record(model, advan):
    """Update $MODEL"""
    newmap = new_compartmental_map(model.statements.ode_system, model._compartment_map)
    if advan in ['ADVAN1', 'ADVAN2', 'ADVAN3', 'ADVAN4', 'ADVAN10', 'ADVAN11', 'ADVAN12']:
        model.control_stream.remove_records(model.control_stream.get_records('MODEL'))
    else:
        oldmap = model._compartment_map
        if oldmap != newmap:
            model.control_stream.remove_records(model.control_stream.get_records('MODEL'))
            mod = model.control_stream.insert_record('$MODEL\n')
            output_name = model.statements.ode_system.output_compartment.name
            comps = {v: k for k, v in newmap.items() if k != output_name}
            i = 1
            while True:
                if i not in comps:
                    break
                if i == 1:
                    mod.add_compartment(comps[i], dosing=True)
                else:
                    mod.add_compartment(comps[i], dosing=False)
                i += 1
    model._compartment_map = newmap


def add_needed_pk_parameters(model, advan, trans):
    """Add missing pk parameters that NONMEM needs"""
    statements = model.statements
    odes = statements.ode_system
    if advan == 'ADVAN2' or advan == 'ADVAN4' or advan == 'ADVAN12':
        if not statements.find_assignment('KA'):
            comp, rate = odes.get_compartment_outflows(odes.find_depot(statements))[0]
            ass = Assignment('KA', rate)
            if rate != ass.symbol:
                statements.insert_before_odes(ass)
                odes.add_flow(odes.find_depot(statements), comp, ass.symbol)
    if advan == 'ADVAN3' and trans == 'TRANS4':
        central = odes.central_compartment
        output = odes.output_compartment
        peripheral = odes.peripheral_compartments[0]
        add_parameters_ratio(model, 'CL', 'V1', central, output)
        add_parameters_ratio(model, 'Q', 'V2', peripheral, central)
        add_parameters_ratio(model, 'Q', 'V1', central, peripheral)
    elif advan == 'ADVAN4':
        central = odes.central_compartment
        output = odes.output_compartment
        peripheral = odes.peripheral_compartments[0]
        if trans == 'TRANS1':
            rate1 = odes.get_flow(central, peripheral)
            add_rate_assignment_if_missing(model, 'K23', rate1, central, peripheral)
            rate2 = odes.get_flow(peripheral, central)
            add_rate_assignment_if_missing(model, 'K32', rate2, peripheral, central)
        if trans == 'TRANS4':
            add_parameters_ratio(model, 'CL', 'V2', central, output)
            add_parameters_ratio(model, 'Q', 'V3', peripheral, central)
    elif advan == 'ADVAN12' and trans == 'TRANS4':
        central = odes.central_compartment
        output = odes.output_compartment
        peripheral1 = odes.peripheral_compartments[0]
        peripheral2 = odes.peripheral_compartments[1]
        add_parameters_ratio(model, 'CL', 'V2', central, output)
        add_parameters_ratio(model, 'Q3', 'V3', peripheral1, central)
        add_parameters_ratio(model, 'Q4', 'V4', peripheral2, central)
    elif advan == 'ADVAN11' and trans == 'TRANS4':
        central = odes.central_compartment
        output = odes.output_compartment
        peripheral1 = odes.peripheral_compartments[0]
        peripheral2 = odes.peripheral_compartments[1]
        add_parameters_ratio(model, 'CL', 'V1', central, output)
        add_parameters_ratio(model, 'Q2', 'V2', peripheral1, central)
        add_parameters_ratio(model, 'Q3', 'V3', peripheral2, central)
    elif advan == 'ADVAN5' or advan == 'ADVAN7':
        newmap = new_compartmental_map(odes, model._compartment_map)
        for source in newmap.keys():
            for dest in newmap.keys():
                if source != dest and source != len(newmap):
                    source_comp = odes.find_compartment(source)
                    dest_comp = odes.find_compartment(dest)
                    rate = odes.get_flow(source_comp, dest_comp)
                    if rate is not None:
                        sn = newmap[source]
                        dn = newmap[dest]
                        if len(str(sn)) > 1 or len(str(dn)) > 1:
                            t = 'T'
                        else:
                            t = ''
                        names = [f'K{sn}{dn}', f'K{sn}T{dn}']
                        if dn == len(newmap):
                            names += [f'K{sn}0', f'K{sn}T0']
                            param = f'K{sn}{t}{0}'
                        else:
                            param = f'K{sn}{t}{dn}'
                        add_rate_assignment_if_missing(
                            model, param, rate, source_comp, dest_comp, synonyms=names
                        )


def add_parameters_ratio(model, numpar, denompar, source, dest):
    statements = model.statements
    if not statements.find_assignment(numpar) or not statements.find_assignment(denompar):
        odes = statements.ode_system
        rate = odes.get_flow(source, dest)
        numer, denom = rate.as_numer_denom()
        par1 = Assignment(numpar, numer)
        par2 = Assignment(denompar, denom)
        if rate != par1.symbol / par2.symbol:
            if not statements.find_assignment(numpar):
                statements.insert_before_odes(par1)
                odes.subs({numer: sympy.Symbol(numpar)})
            if not statements.find_assignment(denompar):
                statements.insert_before_odes(par2)
                odes.subs({denom: sympy.Symbol(denompar)})
        odes.add_flow(source, dest, par1.symbol / par2.symbol)


def define_parameter(model, name, value, synonyms=None):
    """Define a parameter in statments if not defined
    Update if already defined as other value
    return True if new assignment was added
    """
    if synonyms is None:
        synonyms = [name]
    for syn in synonyms:
        ass = model.statements.find_assignment(syn)
        if ass:
            if value != ass.expression and value != symbol(name):
                ass.expression = value
            return False
    new_ass = Assignment(name, value)
    model.statements.insert_before_odes(new_ass)
    return True


def add_rate_assignment_if_missing(model, name, value, source, dest, synonyms=None):
    added = define_parameter(model, name, value, synonyms=synonyms)
    if added:
        model.statements.ode_system.add_flow(source, dest, symbol(name))


def update_abbr_record(model, rv_trans):
    trans = dict()
    if not rv_trans:
        return trans
    # Remove already abbreviated symbols
    # FIXME: Doesn't update if name has changed
    kept = rv_trans.copy()
    abbr_recs = model.control_stream.get_records('ABBREVIATED')
    for rec in abbr_recs:
        for rk, rv in rec.replace.items():
            for tk, tv in rv_trans.items():
                if tv.name == rv:
                    del kept[tk]
    rv_trans = kept
    if not rv_trans:
        return trans

    for rv in model.random_variables:
        rv_symb = symbol(rv.name)
        abbr_pattern = re.match(r'ETA_(\w+)', rv.name)
        if abbr_pattern and '_' not in abbr_pattern.group(1):
            parameter = abbr_pattern.group(1)
            nonmem_name = rv_trans[rv_symb]
            abbr_name = f'ETA({parameter})'
            trans[rv_symb] = symbol(abbr_name)
            abbr_record = f'$ABBR REPLACE {abbr_name}={nonmem_name}\n'
            model.control_stream.insert_record(abbr_record)
        elif not re.match(r'(ETA|EPS)\([0-9]\)', rv.name):
            warnings.warn(
                f'Not valid format of name {rv.name}, falling back to NONMEM name. If custom name, '
                f'follow the format "ETA_X" to get "ETA(X)" in $ABBR.'
            )
    return trans


def update_estimation(model):
    new = model.estimation_steps
    try:
        old = model._old_estimation_steps
    except AttributeError:
        old = []
    if old == new:
        return

    delta = code_record.diff(old, new)
    old_records = model.control_stream.get_records('ESTIMATION')
    i = 0
    new_records = []
    for op, est in delta:
        if op == '+':
            est_code = '$ESTIMATION'
            protected_attributes = []
            if est.method == 'FO':
                method = 'ZERO'
            elif est.method == 'FOCE':
                method = 'COND'
            else:
                method = est.method
            est_code += f' METHOD={method}'
            if est.laplace:
                est_code += ' LAPLACE'
                protected_attributes += ['LAPLACE']
            if est.interaction:
                est_code += ' INTER'
                protected_attributes += ['INTERACTION', 'INTER']
            if est.evaluation:
                if est.method == 'FO' or est.method == 'FOCE':
                    est_code += ' MAXEVAL=0'
                    protected_attributes += ['MAXEVALS', 'MAXEVAL']
                else:
                    est_code += ' EONLY=1'
                    protected_attributes += ['EONLY']
            if est.maximum_evaluations is not None:
                est_code += f' MAXEVAL={est.maximum_evaluations}'
                if set(protected_attributes).intersection({'MAXEVALS', 'MAXEVAL'}):
                    raise ValueError('MAXEVAL already set by evaluation=True')
                protected_attributes += ['MAXEVALS', 'MAXEVAL']
            if est.tool_options:
                option_names = {key for key in est.tool_options.keys()}
                overlapping_attributes = set(protected_attributes).intersection(option_names)
                if overlapping_attributes:
                    overlapping_attributes_str = ', '.join(list(overlapping_attributes))
                    raise ValueError(
                        f'{overlapping_attributes_str} already set as attribute in '
                        f'estimation method object'
                    )
                options_code = ' '.join(
                    [
                        f'{key}={value}'.upper() if value else f'{key}'
                        for key, value in est.tool_options.items()
                    ]
                )
                est_code += f' {options_code}'
            est_code += '\n'
            newrec = create_record(est_code)
            new_records.append(newrec)
        elif op == '-':
            i += 1
        else:
            new_records.append(old_records[i])
            i += 1
    if old_records:
        model.control_stream.replace_records(old_records, new_records)
    else:
        for rec in new_records:
            model.control_stream.insert_record(str(rec))

    old_cov = False
    for est in old:
        old_cov |= est.cov
    new_cov = False
    for est in new:
        new_cov |= est.cov
    if not old_cov and new_cov:
        # Add $COV
        model.control_stream.insert_record('$COVARIANCE\n')
    elif old_cov and not new_cov:
        # Remove $COV
        covrecs = model.control_stream.get_records('COVARIANCE')
        model.control_stream.remove_records(covrecs)

    model._old_estimation_steps = copy.deepcopy(new)


def update_ccontra(model, path=None, force=False):
    h = model.observation_transformation
    y = model.dependent_variable
    dhdy = sympy.diff(h, y)
    ll = -2 * sympy.log(dhdy)
    ll = ll.subs(y, sympy.Symbol('y', real=True, positive=True))
    ll = model.parameters.simplify(ll)
    ll = ll.subs(sympy.Symbol('y', real=True, positive=True), y)

    tr = model.parameter_translation(reverse=True, remove_idempotent=True, as_symbols=True)
    ll = ll.subs(tr)
    h = h.subs(tr)

    # FIXME: break out into method to get path
    if path is None:
        path = Path('.')
    else:
        path = path.parent
    contr_path = path / f'{model.name}_contr.f90'
    ccontr_path = path / f'{model.name}_ccontra.f90'

    contr = """      subroutine contr (icall,cnt,ier1,ier2)
      double precision cnt
      call ncontr (cnt,ier1,ier2,l2r)
      return
      end
"""
    with open(contr_path, 'w') as fh:
        fh.write(contr)

    ccontr1 = """      subroutine ccontr (icall,c1,c2,c3,ier1,ier2)
      USE ROCM_REAL,   ONLY: theta=>THETAC,y=>DV_ITM2
      USE NM_INTERFACE,ONLY: CELS
      double precision c1,c2,c3,w,one,two
      dimension c2(:),c3(:,:)
      if (icall.le.1) return
      w=y(1)

"""

    ccontr2 = """
      call cels (c1,c2,c3,ier1,ier2)
      y(1)=w
"""

    ccontr3 = """
      return
      end
"""

    with open(ccontr_path, 'w') as fh:
        fh.write(ccontr1)
        e1 = sympy.printing.fortran.fcode(h.subs(y, sympy.Symbol('y(1)')), assign_to='y(1)')
        fh.write(e1)
        fh.write(ccontr2)
        e2 = sympy.printing.fortran.fcode(
            sympy.Symbol('c1') + ll.subs(y, sympy.Symbol('y(1)')), assign_to='c1'
        )
        fh.write(e2)
        fh.write(ccontr3)



import pytest

from pharmpy import Model
from pharmpy.modeling import evaluate_expression


def test_evaluate_expression(testdata):
    model = Model(testdata / 'nonmem' / 'models' / 'pheno_noifs.mod')
    ser = evaluate_expression(model, 'TVV')
    assert ser[0] == pytest.approx(1.413062)
    assert ser[743] == pytest.approx(1.110262)



import pytest


@pytest.mark.parametrize(
    "buf,value",
    [
        ('$SIMULATION (1212) NSUBPROBS=0', 1),
        ('$SIMULATION (1212) NSUBPROBLEMS=1', 1),
        ('$SIMULATION (1212) SUBPROB=60', 60),
    ],
)
def test_nsubs(parser, buf, value):
    recs = parser.parse(buf)
    rec = recs.records[0]
    assert rec.nsubs == value



from pathlib import Path

from lark import Lark

from pharmpy.parse_utils import GenericParser

grammar_root = Path(__file__).parent.resolve() / 'grammars'


def install_grammar(cls):
    grammar = Path(grammar_root / cls.grammar_filename).resolve()
    with open(str(grammar), 'r') as fh:
        cls.lark = Lark(fh, **GenericParser.lark_options)
    return cls


class RecordParser(GenericParser):
    pass


@install_grammar
class AbbreviatedRecordParser(RecordParser):
    grammar_filename = 'abbreviated_record.lark'


@install_grammar
class SimulationRecordParser(RecordParser):
    grammar_filename = 'simulation_record.lark'


@install_grammar
class ProblemRecordParser(RecordParser):
    grammar_filename = 'problem_record.lark'
    non_empty = [
        {'root': (0, 'raw_title')},
        {'raw_title': (0, 'REST_OF_LINE')},
    ]


@install_grammar
class ThetaRecordParser(RecordParser):
    grammar_filename = 'theta_record.lark'
    non_empty = [
        {'comment': (1, 'COMMENT')},
    ]


@install_grammar
class OmegaRecordParser(RecordParser):
    grammar_filename = 'omega_record.lark'
    non_empty = [
        {'comment': (1, 'COMMENT')},
    ]


@install_grammar
class OptionRecordParser(RecordParser):
    grammar_filename = 'option_record.lark'


@install_grammar
class DataRecordParser(RecordParser):
    grammar_filename = 'data_record.lark'


@install_grammar
class CodeRecordParser(RecordParser):
    grammar_filename = 'code_record.lark'



"""
:meta private:
"""

import sympy

import pharmpy.model
import pharmpy.symbols as symbols
from pharmpy.parameter import Parameter
from pharmpy.random_variables import RandomVariable
from pharmpy.statements import Assignment


def _preparations(model):
    stats = model.statements
    y = model.dependent_variable
    f = model.statements.find_assignment(y.name).expression
    for eps in model.random_variables.epsilons:
        f = f.subs({symbols.symbol(eps.name): 0})
    return stats, y, f


def remove_error_model(model):
    """Remove error model.

    Parameters
    ----------
    model : Model
        Remove error model for this model

    Return
    ------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import remove_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y := EPS(1)⋅W + F
    >>> remove_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y := F

    Warnings
    --------
    Removing the error model will make the model unrunable for some tools.

    """
    stats, y, f = _preparations(model)
    stats.reassign(y, f)
    model.remove_unused_parameters_and_rvs()
    return model


def set_additive_error_model(model, data_trans=None, series_terms=2):
    r"""Set an additive error model. Initial estimate for new sigma is :math:`(min(DV)/2)²`.

    The error function being applied depends on the data transformation. The table displays
    some examples.

    +------------------------+----------------------------------------+
    | Data transformation    | Additive error                         |
    +========================+========================================+
    | :math:`y`              | :math:`f + \epsilon_1`                 |
    +------------------------+----------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \frac{\epsilon_1}{f}` |
    +------------------------+----------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model. Series expansion will be used for approximation.
    series_terms : int
        Number of terms to use for the series expansion approximation for data
        transformation.

    Return
    ------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import set_additive_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y := EPS(1)⋅W + F
    >>> set_additive_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y := F + εₐ

    >>> from pharmpy.modeling import set_additive_error_model, load_example_model
    >>> model = load_example_model("pheno")
    >>> model.statements.find_assignment("Y")
    Y := EPS(1)⋅W + F
    >>> set_additive_error_model(model, data_trans="log(Y)")    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
                  εₐ
         log(F) + ──
    Y :=          F

    See Also
    --------
    set_proportional_error_model : Proportional error model
    set_combined_error_model : Combined error model

    """
    if has_additive_error_model(model):
        return model
    stats, y, f = _preparations(model)
    ruv = model.create_symbol('epsilon_a')

    data_trans = pharmpy.model.canonicalize_data_transformation(model, data_trans)
    expr = f + ruv
    if data_trans != model.dependent_variable:
        expr = data_trans.subs(model.dependent_variable, expr).series(ruv, n=series_terms).removeO()

    stats.reassign(y, expr)
    model.remove_unused_parameters_and_rvs()

    sigma = model.create_symbol('sigma')
    sigma_par = Parameter(sigma.name, init=_get_prop_init(model.dataset))
    model.parameters.append(sigma_par)

    eps = RandomVariable.normal(ruv.name, 'RUV', 0, sigma)
    model.random_variables.append(eps)
    return model


def _get_prop_init(dt):
    dv_min = dt.pharmpy.observations.min()
    if dv_min == 0:
        return 0.01
    else:
        return (dv_min / 2) ** 2


def set_proportional_error_model(model, data_trans=None):
    r"""Set a proportional error model. Initial estimate for new sigma is 0.09.

    The error function being applied depends on the data transformation.

    +------------------------+----------------------------------------+
    | Data transformation    | Proportional error                     |
    +========================+========================================+
    | :math:`y`              | :math:`f + f \epsilon_1`               |
    +------------------------+----------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \epsilon_1`           |
    +------------------------+----------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model.

    Returns
    -------
    Model
        Reference to the same model object

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_proportional_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y := F⋅εₚ + F

    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_proportional_error_model(model, data_trans="log(Y)")    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y := εₚ + log(F)

    See Also
    --------
    set_additive_error_model : Additive error model
    set_combined_error_model : Combined error model

    """
    if has_proportional_error_model(model):
        return model
    stats, y, f = _preparations(model)
    ruv = model.create_symbol('epsilon_p')

    data_trans = pharmpy.model.canonicalize_data_transformation(model, data_trans)
    if data_trans == sympy.log(model.dependent_variable):
        expr = sympy.log(f) + ruv
    elif data_trans == model.dependent_variable:
        expr = f + f * ruv
    else:
        raise ValueError(f"Not supported data transformation {data_trans}")

    stats.reassign(y, expr)
    model.remove_unused_parameters_and_rvs()

    sigma = model.create_symbol('sigma')
    sigma_par = Parameter(sigma.name, init=0.09)
    model.parameters.append(sigma_par)

    eps = RandomVariable.normal(ruv.name, 'RUV', 0, sigma)
    model.random_variables.append(eps)
    return model


def set_combined_error_model(model, data_trans=None):
    r"""Set a combined error model. Initial estimates for new sigmas are :math:`(min(DV)/2)²` for
    proportional and 0.09 for additive.

    The error function being applied depends on the data transformation.

    +------------------------+-----------------------------------------------------+
    | Data transformation    | Combined error                                      |
    +========================+=====================================================+
    | :math:`y`              | :math:`f + f \epsilon_1 + \epsilon_2`               |
    +------------------------+-----------------------------------------------------+
    | :math:`log(y)`         | :math:`\log(f) + \epsilon_1 + \frac{\epsilon_2}{f}` |
    +------------------------+-----------------------------------------------------+

    Parameters
    ----------
    model : Model
        Set error model for this model
    data_trans : str or expression
        A data transformation expression or None (default) to use the transformation
        specified by the model.

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_combined_error_model(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y := F⋅εₚ + F + εₐ

    >>> from pharmpy.modeling import *
    >>> model = remove_error_model(load_example_model("pheno"))
    >>> set_combined_error_model(model, data_trans="log(Y)")    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
                      εₐ
        εₚ + log(F) + ──
    Y :=              F

    See Also
    --------
    set_additive_error_model : Additive error model
    set_proportional_error_model: Proportional error model

    """
    if has_combined_error_model(model):
        return model
    stats, y, f = _preparations(model)
    ruv_prop = model.create_symbol('epsilon_p')
    ruv_add = model.create_symbol('epsilon_a')

    data_trans = pharmpy.model.canonicalize_data_transformation(model, data_trans)
    if data_trans == sympy.log(model.dependent_variable):
        expr = sympy.log(f) + ruv_prop + ruv_add / f
    elif data_trans == model.dependent_variable:
        expr = f + f * ruv_prop + ruv_add
    else:
        raise ValueError(f"Not supported data transformation {data_trans}")

    stats.reassign(y, expr)
    model.remove_unused_parameters_and_rvs()

    sigma_prop = model.create_symbol('sigma_prop')
    sigma_par1 = Parameter(sigma_prop.name, init=0.09)
    model.parameters.append(sigma_par1)
    sigma_add = model.create_symbol('sigma_add')
    sigma_par2 = Parameter(sigma_add.name, init=_get_prop_init(model.dataset))
    model.parameters.append(sigma_par2)

    eps_prop = RandomVariable.normal(ruv_prop.name, 'RUV', 0, sigma_prop)
    model.random_variables.append(eps_prop)
    eps_add = RandomVariable.normal(ruv_add.name, 'RUV', 0, sigma_add)
    model.random_variables.append(eps_add)
    return model


def has_additive_error_model(model):
    """Check if a model has an additive error model

    Parameters
    ----------
    model : Model
        The model to check

    Return
    ------
    bool
        True if the model has an additive error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_additive_error_model
    >>> model = load_example_model("pheno")
    >>> has_additive_error_model(model)
    False

    See Also
    --------
    has_proportional_error_model : Check if a model has a proportional error model
    has_combined_error_model : Check if a model has a combined error model
    """
    y = model.dependent_variable
    expr = model.statements.after_odes.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {
        symbols.symbol(rv.name) for rv in rvs if symbols.symbol(rv.name) in expr.free_symbols
    }
    if len(rvs_in_y) != 1:
        return False
    eps = rvs_in_y.pop()
    return eps not in (expr - eps).simplify().free_symbols


def has_proportional_error_model(model):
    """Check if a model has a proportional error model

    Parameters
    ----------
    model : Model
        The model to check

    Return
    ------
    bool
        True if the model has a proportional error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_proportional_error_model
    >>> model = load_example_model("pheno")
    >>> has_proportional_error_model(model)
    True

    See Also
    --------
    has_additive_error_model : Check if a model has an additive error model
    has_combined_error_model : Check if a model has a combined error model
    """
    y = model.dependent_variable
    expr = model.statements.after_odes.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {
        symbols.symbol(rv.name) for rv in rvs if symbols.symbol(rv.name) in expr.free_symbols
    }
    if len(rvs_in_y) != 1:
        return False
    eps = rvs_in_y.pop()
    return eps not in (expr / (1 + eps)).simplify().free_symbols


def has_combined_error_model(model):
    """Check if a model has a combined additive and proportinal error model

    Parameters
    ----------
    model : Model
        The model to check

    Return
    ------
    bool
        True if the model has a combined error model and False otherwise

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, has_combined_error_model
    >>> model = load_example_model("pheno")
    >>> has_combined_error_model(model)
    False

    See Also
    --------
    has_additive_error_model : Check if a model has an additive error model
    has_proportional_error_model : Check if a model has a proportional error model
    """
    y = model.dependent_variable
    expr = model.statements.after_odes.full_expression(y)
    rvs = model.random_variables.epsilons
    rvs_in_y = {
        symbols.symbol(rv.name) for rv in rvs if symbols.symbol(rv.name) in expr.free_symbols
    }
    if len(rvs_in_y) != 2:
        return False
    eps1 = rvs_in_y.pop()
    eps2 = rvs_in_y.pop()
    canc1 = ((expr - eps1) / (eps2 + 1)).simplify()
    canc2 = ((expr - eps2) / (eps1 + 1)).simplify()
    return (
        eps1 not in canc1.free_symbols
        and eps2 not in canc1.free_symbols
        or eps1 not in canc2.free_symbols
        and eps2 not in canc2.free_symbols
    )


def use_thetas_for_error_stdev(model):
    """Use thetas to estimate standard deviation of error

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, use_thetas_for_error_stdev
    >>> model = load_example_model("pheno")
    >>> use_thetas_for_error_stdev(model)    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.find_assignment("Y")
    Y := EPS(1)⋅SD_EPS(1)⋅W + F

    See also
    --------
    set_weighted_error_model : Encode error model with one epsilon and weight
    """
    rvs = model.random_variables.epsilons
    for eps in rvs:
        sigmas = eps.parameter_names
        if len(sigmas) > 1:
            raise ValueError('use_thetas_for_error_stdev only supports non-correlated sigmas')
        sigma = sigmas[0]
        param = model.parameters[sigma]
        param.fix = True
        theta_init = param.init ** 0.5
        param.init = 1
        theta = Parameter(f'SD_{eps.name}', theta_init, lower=0)
        model.parameters.append(theta)
        symb = sympy.Symbol(eps.name)
        model.statements.subs({symb: theta.symbol * symb})
    return model


def set_weighted_error_model(model):
    """Encode error model with one epsilon and W as weight

    Parameters
    ----------
    model : Model
        Pharmpy model

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_weighted_error_model
    >>> model = load_example_model("pheno")
    >>> set_weighted_error_model(model)    # doctest: +ELLIPSIS
    <...>

    See also
    --------
    use_thetas_for_error_stdev : Use thetas to estimate error
    """
    stats, y, f = _preparations(model)
    epsilons = model.random_variables.epsilons
    expr = stats.find_assignment(y.name).expression
    ssum = 0
    q = sympy.Q.real(y)  # Dummy predicate
    for term in expr.args:
        eps = [x for x in term.free_symbols if x.name in epsilons.names]
        if len(eps) > 0:
            eps = eps[0]
            remaining = term / eps
            ssum += remaining ** 2
            for symb in remaining.free_symbols:
                q &= sympy.Q.positive(symb)
    w = sympy.sqrt(ssum)
    w = sympy.refine(w, q)

    for i, s in enumerate(stats):
        if isinstance(s, Assignment) and s.symbol == y:
            stats.insert(i, Assignment('W', w))
            break

    stats.reassign(y, f + sympy.Symbol('W') * sympy.Symbol(epsilons[0].name))
    model.remove_unused_parameters_and_rvs()
    return model


def set_dtbs_error_model(model, fix_to_log=False):
    """Dynamic transform both sides

    Parameters
    ----------
    model : Model
        Pharmpy model
    fix_to_log : Boolean
        Set to True to fix lambda and zeta to 0, i.e. emulating log-transformed data

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, set_dtbs_error_model
    >>> model = load_example_model("pheno")
    >>> set_dtbs_error_model(model)    # doctest: +ELLIPSIS
    <...>

    """
    use_thetas_for_error_stdev(model)
    set_weighted_error_model(model)
    stats, y, f = _preparations(model)
    tbs_lambda = Parameter('tbs_lambda', 1)
    tbs_zeta = Parameter('tbs_zeta', 0.001)
    if fix_to_log:
        tbs_lambda.fix = True
        tbs_lambda.init = 0
        tbs_zeta.fix = True
        tbs_zeta.init = 0
    model.parameters.append(tbs_lambda)
    model.parameters.append(tbs_zeta)
    lam = tbs_lambda.symbol
    zeta = tbs_zeta.symbol

    for i, s in enumerate(stats):
        if isinstance(s, Assignment) and s.symbol == sympy.Symbol('W'):
            break

    stats.insert(i + 1, Assignment('W', (f ** zeta) * sympy.Symbol('W')))
    ipred = sympy.Piecewise(
        ((f ** lam - 1) / lam, sympy.And(sympy.Ne(lam, 0), sympy.Ne(f, 0))),
        (sympy.log(f), sympy.And(sympy.Eq(lam, 0), sympy.Ne(f, 0))),
        (-1 / lam, sympy.And(sympy.Eq(lam, 0), sympy.Eq(f, 0))),
        (-1000000000, True),
    )
    stats.insert(i + 2, Assignment('IPRED', ipred))
    yexpr = stats.find_assignment(y.name)
    yexpr.subs({f: sympy.Symbol('IPRED')})

    obs = sympy.Piecewise(
        (sympy.log(y), sympy.Eq(lam, 0)), ((y ** lam - 1) / lam, sympy.Ne(lam, 0))
    )
    model.observation_transformation = obs

    return model



import altair as alt

import pharmpy.visualization
from pharmpy.data import PharmDataFrame


def plot_iofv_vs_iofv(model, other):
    """Plot individual OFV of two models against each other

    Parameters
    ----------
    model : Model
        The first model
    other : Model
        The second model

    Results
    -------
    alt.Chart
        Scatterplot

    """
    x_label = f'{model.name} iOFV'
    y_label = f'{other.name} iOFV'
    df = PharmDataFrame(
        {
            x_label: model.modelfit_results.individual_ofv,
            y_label: other.modelfit_results.individual_ofv,
        }
    )
    id_name = df.index.name
    df = df.reset_index()
    plot = pharmpy.visualization.scatter_plot_correlation(
        df, x_label, y_label, tooltip_columns=[id_name], title='iOFV vs iOFV'
    )
    return plot


def plot_individual_predictions(model, predictions=None, individuals=None):
    """Plot DV and predictions grouped on individuals

    Parameters
    ----------
    model : Model
        Previously run Pharmpy model.
    predictions : list
        A list of names of predictions to plot. None for all available
    individuals : list
        A list of individuals to include. None for all individuals

    Returns
    -------
    alt.Chart
        Plot

    """
    res = model.modelfit_results
    pred = res.predictions
    obs = model.dataset.pharmpy.observations
    indexcols = pred.index.names
    idcol = indexcols[0]
    idvcol = indexcols[1]

    data = pred.join(obs).reset_index()
    data = data.melt(id_vars=indexcols)

    if individuals is not None:
        data = data[data[idcol].isin(individuals)]
    if predictions is not None:
        dvcol = obs.name
        data = data[data['variable'].isin(predictions + [dvcol])]

    plot = (
        alt.Chart(data)
        .mark_line(point=True)
        .encode(x=idvcol, y='value', color='variable', tooltip=[idvcol, 'value'])
        .facet(f'{idcol}:N', columns=5)
    )
    return plot



from pathlib import Path

import pytest
from pyfakefs.fake_filesystem_unittest import Patcher

# Do not let pyfakefs patch mpmath since it emits warnings
# and doesn't do anything with the filesystem anyway
Patcher.SKIPNAMES.update(['mpmath'])


@pytest.fixture(scope='session')
def testdata():
    """Test data (root) folder."""
    return Path(__file__).resolve().parent / 'testdata'


@pytest.fixture(scope='session')
def datadir(testdata):
    return testdata / 'nonmem'


@pytest.fixture(scope='session')
def pheno_path(datadir):
    return datadir / 'pheno_real.mod'



"""Utils for managing plugins."""

import importlib
import pkgutil
from pathlib import Path


class PluginError(Exception):
    pass


def detect_model(src):
    """Detects appropriate implementation from a source object
    Return a model object
    """

    plugins = load_plugins()
    detected_classes = []
    for module in plugins:
        if hasattr(module, 'detect_model'):
            cls = module.detect_model(src)
            if cls:
                detected_classes.append(cls)

    if len(detected_classes) == 0:
        raise PluginError(f"No support for model {src}")
    if len(detected_classes) > 1:
        raise PluginError(f"More than one model plugin supports model {src}")
    return detected_classes[0]


def load_plugins():
    """Find and import all available plugins"""
    plugin_path = Path(__file__).resolve().parent

    modules = list()

    # str on path to workaround https://bugs.python.org/issue44061
    for _, modname, ispkg in pkgutil.iter_modules([str(plugin_path)], 'pharmpy.plugins.'):
        if ispkg:
            module = importlib.import_module(modname)
            modules.append(module)

    return modules



from operator import add, mul

import pytest
import sympy

from pharmpy.modeling.eta_additions import EtaAddition
from pharmpy.symbols import symbol as S


@pytest.mark.parametrize(
    'addition,expression',
    [
        (EtaAddition.exponential(add), S('CL') + sympy.exp(S('eta_new'))),
        (EtaAddition.exponential(mul), S('CL') * sympy.exp(S('eta_new'))),
    ],
)
def test_apply(addition, expression):
    addition.apply(original='CL', eta='eta_new')

    assert addition.template == expression



"""
:meta private:
"""

import re

import sympy
from sympy import exp, sign

from pharmpy.modeling.help_functions import _format_input_list, _get_etas
from pharmpy.parameter import Parameter
from pharmpy.statements import Assignment, ModelStatements
from pharmpy.symbols import symbol as S


def transform_etas_boxcox(model, list_of_etas=None):
    """Applies a boxcox transformation to selected etas

    Initial estimate for lambda is 0.1 with bounds (-3, 3).

    Parameters
    ----------
    model : Model
        Pharmpy model to apply boxcox transformation to.
    list_of_etas : str, list
        Name/names of etas to transform. If None, all etas will be transformed (default).

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> transform_etas_boxcox(model, ["ETA(1)"])    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.before_odes.full_expression("CL")
    THETA(1)*WGT*exp((exp(ETA(1))**lambda1 - 1)/lambda1)

    See also
    --------
    transform_etas_tdist
    transform_etas_john_draper

    """
    list_of_etas = _format_input_list(list_of_etas)
    etas = _get_etas(model, list_of_etas)
    eta_transformation = EtaTransformation.boxcox(len(etas))
    _transform_etas(model, eta_transformation, etas)
    return model


def transform_etas_tdist(model, list_of_etas=None):
    """Applies a t-distribution transformation to selected etas

    Initial estimate for degrees of freedom is 80 with bounds (3, 100).

    Parameters
    ----------
    model : Model
        Pharmpy model to apply t distribution transformation to.
    list_of_etas : str, list
        Name/names of etas to transform. If None, all etas will be transformed (default).

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> transform_etas_tdist(model, ["ETA(1)"])    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.before_odes.full_expression("CL")    # doctest: +ELLIPSIS
    THETA(1)*WGT*exp(ETA(1)*(1 + (ETA(1)**2 + 1)/(4*df1) + (5*ETA(1)**4 + 16*ETA(1)**2 + 3)/(96*...

    See also
    --------
    transform_etas_boxcox
    transform_etas_john_draper

    """
    list_of_etas = _format_input_list(list_of_etas)
    etas = _get_etas(model, list_of_etas)
    eta_transformation = EtaTransformation.tdist(len(etas))
    _transform_etas(model, eta_transformation, etas)
    return model


def transform_etas_john_draper(model, list_of_etas=None):
    """Applies a John Draper transformation [1]_ to spelected etas

    Initial estimate for lambda is 0.1 with bounds (-3, 3).

    .. [1] John, J., Draper, N. (1980). An Alternative Family of Transformations.
       Journal of the Royal Statistical Society. Series C (Applied Statistics),
       29(2), 190-197. doi:10.2307/2986305

    Parameters
    ----------
    model : Model
        Pharmpy model to apply John Draper transformation to.
    list_of_etas : str, list
        Name/names of etas to transform. If None, all etas will be transformed (default).

    Return
    ------
    Model
        Reference to the same model

    Examples
    --------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> transform_etas_john_draper(model, ["ETA(1)"])    # doctest: +ELLIPSIS
    <...>
    >>> model.statements.before_odes.full_expression("CL")
    THETA(1)*WGT*exp(((Abs(ETA(1)) + 1)**lambda1 - 1)*sign(ETA(1))/lambda1)

    See also
    --------
    transform_etas_boxcox
    transform_etas_tdist

    """
    list_of_etas = _format_input_list(list_of_etas)
    etas = _get_etas(model, list_of_etas)
    eta_transformation = EtaTransformation.john_draper(len(etas))
    _transform_etas(model, eta_transformation, etas)
    return model


def _transform_etas(model, transformation, etas):
    etas_assignment, etas_subs = _create_new_etas(etas, transformation.name)
    thetas = _create_new_thetas(model, transformation.theta_type, len(etas))
    transformation.apply(etas_assignment, thetas)
    statements_new = transformation.assignments
    sset = model.statements
    sset.subs(etas_subs)

    statements_new.extend(sset)

    model.statements = statements_new


def _create_new_etas(etas_original, transformation):
    etas_subs = dict()
    etas_assignment = dict()
    if transformation == 'boxcox':
        eta_new = 'etab'
    elif transformation == 'tdist':
        eta_new = 'etat'
    elif transformation == 'johndraper':
        eta_new = 'etad'
    else:
        eta_new = 'etan'
    for i, eta in enumerate(etas_original, 1):
        etas_subs[eta.symbol] = sympy.Symbol(f'{eta_new.upper()}{i}')
        etas_assignment[sympy.Symbol(f'{eta_new}{i}')] = sympy.Symbol(f'{eta_new.upper()}{i}')
        etas_assignment[sympy.Symbol(f'eta{i}')] = eta.symbol

    return etas_assignment, etas_subs


def _create_new_thetas(model, transformation, no_of_thetas):
    pset = model.parameters
    thetas = dict()
    theta_name = str(model.create_symbol(stem=transformation, force_numbering=True))

    if transformation == 'lambda':
        param_settings = [0.01, -3, 3]
    else:
        param_settings = [80, 3, 100]

    if no_of_thetas == 1:
        pset.append(Parameter(theta_name, *param_settings))
        thetas['theta1'] = theta_name
    else:
        theta_no = int(re.findall(r'\d', theta_name)[0])

        for i in range(1, no_of_thetas + 1):
            pset.append(Parameter(theta_name, 0.01, -3, 3))
            thetas[f'theta{i}'] = theta_name
            theta_name = f'{transformation}{theta_no + i}'

    model.parameters = pset

    return thetas


class EtaTransformation:
    def __init__(self, name, assignments, theta_type):
        self.name = name
        self.assignments = assignments
        self.theta_type = theta_type

    def apply(self, etas, thetas):
        for assignment in self.assignments:
            assignment.subs(etas)
            assignment.subs(thetas)

    @classmethod
    def boxcox(cls, no_of_etas):
        assignments = ModelStatements()
        for i in range(1, no_of_etas + 1):
            symbol = S(f'etab{i}')
            expression = (exp(S(f'eta{i}')) ** S(f'theta{i}') - 1) / (S(f'theta{i}'))

            assignment = Assignment(symbol, expression)
            assignments.append(assignment)

        return cls('boxcox', assignments, 'lambda')

    @classmethod
    def tdist(cls, no_of_etas):
        assignments = ModelStatements()
        for i in range(1, no_of_etas + 1):
            symbol = S(f'etat{i}')

            eta = S(f'eta{i}')
            theta = S(f'theta{i}')

            num_1 = eta ** 2 + 1
            denom_1 = 4 * theta

            num_2 = (5 * eta ** 4) + (16 * eta ** 2 + 3)
            denom_2 = 96 * theta ** 2

            num_3 = (3 * eta ** 6) + (19 * eta ** 4) + (17 * eta ** 2) - 15
            denom_3 = 384 * theta ** 3

            expression = eta * (1 + (num_1 / denom_1) + (num_2 / denom_2) + (num_3 / denom_3))

            assignment = Assignment(symbol, expression)
            assignments.append(assignment)

        return cls('tdist', assignments, 'df')

    @classmethod
    def john_draper(cls, no_of_etas):
        assignments = ModelStatements()
        for i in range(1, no_of_etas + 1):
            symbol = S(f'etad{i}')

            eta = S(f'eta{i}')
            theta = S(f'theta{i}')

            expression = sign(eta) * (((abs(eta) + 1) ** theta - 1) / theta)

            assignment = Assignment(symbol, expression)
            assignments.append(assignment)

        return cls('johndraper', assignments, 'lambda')

    def __str__(self):
        return str(self.assignments)



from pharmpy import Model
from pharmpy.plugins.nlmixr import convert_model


def test_model(testdata):
    nmmodel = Model(testdata / 'nonmem' / 'pheno.mod')
    model = convert_model(nmmodel)
    assert 'ini' in str(model)


def test_pheno_real(testdata):
    nmmodel = Model(testdata / 'nonmem' / 'pheno_real.mod')
    model = convert_model(nmmodel)
    print(str(model))
    assert '} else {' in str(model)



import numpy as np
import symengine
import sympy


def evaluate_expression(model, expression):
    """Evaluate expression using model

    Calculate the value of expression for each data record.
    The expression can contain dataset columns, variables in model and
    population parameters. If the model has parameter estimates these
    will be used. Initial estimates will be used for non-estimated parameters.

    Parameters
    ----------
    model : Model
        Pharmpy model
    expression : str or sympy expression
        Expression to evaluate

    Returns
    -------
    pd.Series
        A series of one evaluated value for each data record

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model, evaluate_expression
    >>> model = load_example_model("pheno")
    >>> evaluate_expression(model, "TVCL*1000")
    0      6.573770
    1      6.573770
    2      6.573770
    3      6.573770
    4      6.573770
             ...
    739    5.165105
    740    5.165105
    741    5.165105
    742    5.165105
    743    5.165105
    Length: 744, dtype: float64

    """
    expression = sympy.sympify(expression)
    full_expr = model.statements.before_odes.full_expression(expression)
    pe = model.modelfit_results.parameter_estimates
    inits = model.parameters.inits
    expr = full_expr.subs(dict(pe)).subs(inits)
    data = model.dataset
    expr = symengine.sympify(expr)

    def func(row):
        subs = expr.subs(dict(row))
        return np.float64(subs.evalf())

    df = data.apply(func, axis=1)
    return df



import pytest


@pytest.mark.parametrize(
    "buf,option,old_value,value,expected",
    [
        ('$SIZES PC=23', 'PC', 23, 40, '$SIZES PC=40'),
        ('$SIZES LIM1=5', 'PC', 30, 23, '$SIZES LIM1=5'),
        ('$SIZES LIM1=5', 'PC', 30, 31, '$SIZES LIM1=5 PC=31'),
        ('$SIZES LIM1=5 PC=23', 'PC', 23, 29, '$SIZES LIM1=5'),
    ],
)
def test_set_option(parser, buf, option, old_value, value, expected):
    recs = parser.parse(buf)
    rec = recs.records[0]
    assert getattr(rec, option) == old_value
    setattr(rec, option, value)
    assert str(rec) == expected


@pytest.mark.parametrize(
    "buf,option,value,ex",
    [
        ('$SIZES PC=23', 'PC', 100, ValueError),
    ],
)
def test_exceptions(parser, buf, option, value, ex):
    recs = parser.parse(buf)
    rec = recs.records[0]
    with pytest.raises(ex):
        setattr(rec, option, value)



import pandas as pd
import pytest

import pharmpy.plugins.nonmem.table as table


def test_nonmem_table(pheno_ext):
    ext_table_file = table.NONMEMTableFile(pheno_ext)
    assert list(ext_table_file.table.data_frame.columns) == [
        'ITERATION',
        'THETA(1)',
        'THETA(2)',
        'THETA(3)',
        'OMEGA(1,1)',
        'OMEGA(2,1)',
        'OMEGA(2,2)',
        'SIGMA(1,1)',
        'OBJ',
    ]


def test_ext_table(pheno_ext):
    ext_table_file = table.NONMEMTableFile(pheno_ext)
    ext_table = ext_table_file.table
    assert ext_table.number == 1
    assert ext_table.is_evaluation is False
    assert ext_table.problem == 1
    assert ext_table.subproblem == 0
    assert ext_table.iteration2 == 0
    assert ext_table.method == "First Order Conditional Estimation with Interaction"
    assert ext_table.goal_function == "MINIMUM VALUE OF OBJECTIVE FUNCTION"
    assert ext_table.final_parameter_estimates['THETA(1)'] == 0.00469555
    assert list(ext_table.fixed) == [False, False, False, False, True, False, False]
    assert ext_table.final_ofv == 586.27605628188053
    assert ext_table.initial_ofv == 587.36644134661617


def test_phi_table(pheno_phi):
    phi_table_file = table.NONMEMTableFile(pheno_phi)
    phi_table = phi_table_file.table
    assert phi_table.problem == 1
    assert phi_table.subproblem == 0
    assert phi_table.iteration2 == 0
    assert phi_table.method == "First Order Conditional Estimation with Interaction"
    assert phi_table.goal_function is None


def test_cov_table(pheno_cov):
    cov_table_file = table.NONMEMTableFile(pheno_cov)
    cov_table = cov_table_file.table
    df = cov_table.data_frame
    paramnames = ['THETA(1)', 'THETA(2)', 'THETA(3)', 'OMEGA(1,1)', 'OMEGA(2,2)', 'SIGMA(1,1)']
    assert list(df.columns) == paramnames
    assert list(df.index) == paramnames
    assert df.values[0][0] == 4.41151e-8
    assert df.values[0][3] == -1.09343e-6


def test_create_phi_table(fs):
    df = pd.DataFrame({'ETA(1)': [1, 2], 'ETA(2)': [5, 6]}, index=[1, 2])
    df.index.name = 'ID'
    phi = table.PhiTable(df=df)
    phi.create_content()
    correct = ''' SUBJECT_NO   ID           ETA(1)       ETA(2)      
            1            1  1.00000E+00  5.00000E+00
            2            2  2.00000E+00  6.00000E+00
'''  # noqa W291
    assert phi.content == correct

    table_file = table.NONMEMTableFile(tables=[phi])
    table_file.write('my.phi')
    with open('my.phi', 'r') as fp:
        cont = fp.read()
    correct = 'TABLE NO.     1\n' + correct
    assert cont == correct


def test_errors(testdata):
    ext_file = table.NONMEMTableFile(testdata / 'nonmem' / 'errors' / 'no_header_error.ext')
    with pytest.raises(ValueError):
        ext_file.table_no(2).data_frame



import re
import warnings

from pharmpy.parameter import Parameter, Parameters
from pharmpy.parse_utils.generic import AttrToken, remove_token_and_space

from .record import Record

max_upper_bound = 1000000
min_lower_bound = -1000000


class ThetaRecord(Record):
    def __init__(self, content, parser_class):
        super().__init__(content, parser_class)
        self.name_map = None

    def add_nonmem_name(self, name_original, theta_number):
        if not re.match(r'THETA\(\d+\)', name_original):
            self.root.add_comment_node(name_original)
            self.root.add_newline_node()
        self.name_map = {name_original: theta_number}

    def parameters(self, first_theta, seen_labels=None):
        """Get a parameter set for this theta record.
        first_theta is the number of the first theta in this record
        """
        if seen_labels is None:
            seen_labels = set()
        pset = Parameters()
        current_theta = first_theta
        for theta in self.root.all('theta'):
            init = theta.init.tokens[0].eval
            fix = bool(theta.find('FIX'))
            if theta.find('low'):
                if theta.low.find('NEG_INF'):
                    lower = min_lower_bound
                else:
                    lower = theta.low.tokens[0].eval
            else:
                lower = min_lower_bound
            if theta.find('up'):
                if theta.up.find('POS_INF'):
                    upper = max_upper_bound
                else:
                    upper = theta.up.tokens[0].eval
            else:
                upper = max_upper_bound
            multiple = theta.find('n')
            if multiple:
                n = multiple.INT
            else:
                n = 1
            for i in range(0, n):
                name = None
                import pharmpy.plugins.nonmem as nonmem

                if 'comment' in nonmem.conf.parameter_names:
                    # needed to avoid circular import with Python 3.6
                    found = False
                    for subnode in self.root.tree_walk():
                        if id(subnode) == id(theta):
                            if found:
                                break
                            else:
                                found = True
                                continue
                        if found and subnode.rule == 'NEWLINE':
                            m = re.search(r';\s*([a-zA-Z_]\w*)', str(subnode))
                            if m:
                                name = m.group(1)
                                break
                    if name in seen_labels:
                        warnings.warn(
                            f'The parameter name {name} is duplicated. Falling back to basic '
                            f'NONMEM names.'
                        )
                        name = None

                if not name:
                    name = f'THETA({current_theta})'
                seen_labels.add(name)
                new_par = Parameter(name, init, lower, upper, fix)
                current_theta += 1
                pset.append(new_par)

        if not self.name_map:
            self.name_map = {name: first_theta + i for i, name in enumerate(pset.names)}
        return pset

    def _multiple(self, theta):
        """Return the multiple (xn) of a theta or 1 if no multiple"""
        multiple = theta.find('n')
        if multiple:
            n = multiple.INT
        else:
            n = 1
        return n

    def update(self, parameters, first_theta):
        """From a Parameters update the THETAs in this record

        Currently only updating initial estimates
        """
        i = first_theta
        for theta in self.root.all('theta'):
            name = {v: k for k, v in self.name_map.items()}[i]
            param = parameters[name]
            new_init = param.init
            if float(str(theta.init)) != new_init:
                theta.init.tokens[0].value = str(new_init)
            fix = bool(theta.find('FIX'))
            if fix != param.fix:
                if param.fix:
                    space = AttrToken('WS', ' ')
                    fix_token = AttrToken('FIX', 'FIX')
                    theta.children.extend([space, fix_token])
                else:
                    remove_token_and_space(theta, 'FIX')

            n = self._multiple(theta)
            i += n

    def update_name_map(self, trans):
        """Update name_map given dict from -> to"""
        for key, value in trans.items():
            if key in self.name_map:
                n = self.name_map[key]
                del self.name_map[key]
                self.name_map[value] = n

    def renumber(self, new_start):
        old_start = min(self.name_map.values())
        if new_start != old_start:
            for name in self.name_map:
                self.name_map[name] += new_start - old_start

    def remove(self, names):
        first_theta = min(self.name_map.values())
        indices = {self.name_map[name] - first_theta for name in names}
        for name in names:
            del self.name_map[name]
        keep = []
        i = 0
        for node in self.root.children:
            if node.rule == 'theta':
                if i not in indices:
                    keep.append(node)
                i += 1
            else:
                keep.append(node)
        self.root.children = keep

    def __len__(self):
        """Number of thetas in this record"""
        tot = 0
        for theta in self.root.all('theta'):
            tot += self._multiple(theta)
        return tot



import pytest

import pharmpy.plugins.utils as utils


def test_detect_model():
    with pytest.raises(utils.PluginError):
        utils.detect_model(None)


def test_load_plugins():
    plugins = utils.load_plugins()
    names = [x.__name__ for x in plugins]
    assert set(names) == {
        'pharmpy.plugins.nonmem',
        'pharmpy.plugins.nlmixr',
        'pharmpy.plugins.fcon',
    }



"""
NONMEM data record class.
"""

from pharmpy.model import ModelSyntaxError
from pharmpy.parse_utils import AttrToken, AttrTree

from .option_record import OptionRecord


class DataRecord(OptionRecord):
    @property
    def filename(self):
        """The (raw, unresolved) path of the dataset."""
        filename = self.root.filename
        if filename.find('TEXT'):
            return str(filename)
        else:  # 'QUOTE'
            return str(filename)[1:-1]

    @filename.setter
    def filename(self, value):
        if not value:
            # erase and replace by * (for previous subproblem)
            new = [AttrToken('ASTERISK', '*')]
            nodes = []
            for child in self.root.children:
                if new and child.rule == 'ws':
                    nodes += [child, new.pop()]
                elif child.rule in {'ws', 'comment'}:
                    nodes += [child]
            self.root = AttrTree.create('root', nodes)
        else:
            # replace only 'filename' rule and quote appropriately if, but only if, needed
            filename = str(value)
            quoted = [
                ',',
                ';',
                '(',
                ')',
                '=',
                ' ',
                'IGNORE',
                'NULL',
                'ACCEPT',
                'NOWIDE',
                'WIDE',
                'CHECKOUT',
                'RECORDS',
                'RECORDS',
                'LRECL',
                'NOREWIND',
                'REWIND',
                'NOOPEN',
                'LAST20',
                'TRANSLATE',
                'BLANKOK',
                'MISDAT',
            ]
            if not any(x in filename for x in quoted):
                node = AttrTree.create('filename', {'TEXT': filename})
            else:
                if "'" in filename:
                    node = AttrTree.create('filename', {'QUOTE': '"%s"' % filename})
                else:
                    node = AttrTree.create('filename', {'QUOTE': "'%s'" % filename})
            (pre, old, post) = self.root.partition('filename')
            self.root.children = pre + [node] + post

    @property
    def ignore_character(self):
        """The comment character from ex IGNORE=C or None if not available."""
        if hasattr(self.root, 'ignchar') and self.root.ignchar.find('char'):
            char = str(self.root.ignchar.char)
            if len(char) == 3:  # It must be quoted
                char = char[1:-1]
            return char
        else:
            return None

    @ignore_character.setter
    def ignore_character(self, c):
        if c != self.ignore_character:
            self.root.remove('ignchar')
            char_node = AttrTree.create('char', [{'CHAR': c}])
            node = AttrTree.create('ignchar', [{'IGNORE': 'IGNORE'}, {'EQUALS': '='}, char_node])
            self.append_option_node(node)

    def ignore_character_from_header(self, label):
        """Set ignore character from a header label
        If s[0] is a-zA-Z set @
        else set s[0]
        """
        c = label[0]
        if c.isalpha():
            self.ignore_character = '@'
        else:
            self.ignore_character = c

    @property
    def null_value(self):
        """The value to replace for NULL (i.e. . etc) in the dataset
        note that only +,-,0 (meaning 0) and 1-9 are allowed
        """
        if hasattr(self.root, 'null') and self.root.null.find('char'):
            char = str(self.root.null.char)
            if char == '+' or char == '-':
                return 0
            else:
                return float(char)
        else:
            return 0

    @property
    def ignore(self):
        filters = []
        for option in self.root.all('ignore'):
            for filt in option.all('filter'):
                filters.append(filt)
        return filters

    @ignore.deleter
    def ignore(self):
        self.root.remove('ignore')

    @property
    def accept(self):
        filters = []
        for option in self.root.all('accept'):
            for filt in option.all('filter'):
                filters.append(filt)
        return filters

    @accept.deleter
    def accept(self):
        self.root.remove('accept')

    def validate(self):
        """Syntax validation of this data record
        Assumes only on $DATA exists in this $PROBLEM.
        """
        if len(self.root.all('ignchar')) > 1:
            raise ModelSyntaxError('More than one IGNORE=c')



from pharmpy import Model
from pharmpy.plugins.nonmem.results import simfit_results
from pharmpy.results import Results


class SimfitResults(Results):
    """Simfit results class"""

    def __init__(self):
        pass


def calculate_results(modelfit_results):
    """Calculate simfit results"""
    res = SimfitResults()
    res.modelfit_results = modelfit_results
    return res


def psn_simfit_results(paths):
    modelfit_results = []
    for path in paths:
        model = Model(path)
        modelfit_results.extend(simfit_results(model))
    res = calculate_results(modelfit_results)
    return res



"""
:meta private:
"""

import warnings

from pharmpy.random_variables import RandomVariables


def remove_iov(model):
    """Removes all IOV etas

    Parameters
    ----------
    model : Model
        Pharmpy model to remove IOV from.

    Return
    ------
    Model
        Reference to the same model

    Example
    -------
    >>> from pharmpy.modeling import *
    >>> model = load_example_model("pheno")
    >>> remove_iov(model)       # doctest: +ELLIPSIS
    <...>

    See also
    --------
    add_iiv
    add_iov
    remove_iiv

    """
    rvs, sset = model.random_variables, model.statements
    etas = _get_etas(rvs)

    if not etas:
        warnings.warn('No IOVs present')
        return model

    for eta in etas:
        sset.subs({eta.symbol: 0})
        del rvs[eta]

    model.random_variables = rvs
    model.statements = sset

    model.modelfit_results = None

    return model


def _get_etas(rvs):
    etas = [eta for eta in rvs if eta.level == 'IOV']

    return RandomVariables(etas)
